r"""
Interval Exchange Transformations and Linear Involution

An interval exchage transformation is a map defined on an interval (see
help(iet.IntervalExchangeTransformation) for a more complete help.

EXAMPLES::

    sage: from surface_dynamics import *

Initialization of a simple iet with integer lengths::

    sage: T = iet.IntervalExchangeTransformation(Permutation([3,2,1]), [3,1,2])
    sage: T
    Interval exchange transformation of [0, 6[ with permutation
    1 2 3
    3 2 1

Rotation corresponds to iet with two intervals::

    sage: p = iet.Permutation('a b', 'b a')
    sage: T = iet.IntervalExchangeTransformation(p, [1, (sqrt(5)-1)/2])
    sage: T.in_which_interval(0)
    'a'
    sage: T.in_which_interval(T(0))
    'a'
    sage: T.in_which_interval(T(T(0)))
    'b'
    sage: T.in_which_interval(T(T(T(0))))
    'a'

There are two plotting methods for iet::

    sage: p = iet.Permutation('a b c','c b a')
    sage: T = iet.IntervalExchangeTransformation(p, [1, 2, 3])

.. plot the domain and the range of T::

    sage: T.plot_two_intervals()   # not tested (problem with matplotlib font cache)
    Graphics object consisting of 12 graphics primitives

.. plot T as a function::

    sage: T.plot_function()  # not tested (problem with matplotlib font cache)
    Graphics object consisting of 3 graphics primitives
"""
#*****************************************************************************
#       Copyright (C) 2019 Vincent Delecroix <20100.delecroix@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
#*****************************************************************************

from __future__ import print_function, absolute_import
from six.moves import range, map, filter, zip

from copy import copy
from sage.modules.free_module_element import free_module_element
from sage.rings.all import ZZ, QQ

from .template import side_conversion, interval_conversion
from .labelled import LabelledPermutationIET

def wedge(u, v):
    r"""
    Return the wedge of the vectors ``u`` and ``v``
    """
    d = len(u)
    R = u.base_ring()
    assert len(u) == len(v) and v.base_ring() == R
    return free_module_element(R, d*(d-1)//2, [(u[i]*v[j] - u[j]*v[i]) for i in range(d-1) for j in range(i+1,d)])

class IntervalExchangeTransformation(object):
    r"""
    Interval exchange transformation

    INPUT:

    - ``permutation`` - a permutation (LabelledPermutationIET)

    - ``lengths`` - the list of lengths

    EXAMPLES::

        sage: from surface_dynamics import *

    Direct initialization::

        sage: p = iet.IET(('a b c','c b a'),{'a':1,'b':1,'c':1})
        sage: p.permutation()
        a b c
        c b a
        sage: p.lengths()
        (1, 1, 1)

    Initialization from a iet.Permutation::

        sage: perm = iet.Permutation('a b c','c b a')
        sage: l = vector([0.5,1,1.2])
        sage: t = iet.IET(perm,l)
        sage: t.permutation() == perm
        True
        sage: t.lengths() == l
        True

    Initialization from a Permutation::

        sage: p = Permutation([3,2,1])
        sage: iet.IET(p, [1,1,1])
        Interval exchange transformation of [0, 3[ with permutation
        1 2 3
        3 2 1

    If it is not possible to convert lengths to real values an error is raised::

        sage: iet.IntervalExchangeTransformation(('a b','b a'),['e','f'])
        Traceback (most recent call last):
        ...
        TypeError: unable to convert x (='e') into a real number

    The value for the lengths must be positive::

        sage: iet.IET(('a b','b a'),[-1,-1])
        Traceback (most recent call last):
        ...
        ValueError: lengths must be non-negative
    """
    def __init__(self, permutation=None, lengths=None, base_ring=None):
        r"""
        INPUT:

        - ``permutation`` - a permutation (LabelledPermutationIET)

        - ``lengths`` - the list of lengths

        TEST::

            sage: from surface_dynamics import *

            sage: p = iet.IntervalExchangeTransformation(('a','a'), [1])
            sage: p == loads(dumps(p))
            True
        """
        from sage.modules.free_module_element import free_module_element as vector

        if permutation is None or lengths is None:
            self._permutation = None
            self._lengths = []
            self._base_ring = None
            self._vector_space = None
        else:
            self._permutation = copy(permutation)
            if isinstance(lengths, dict):
                l = [0] * len(permutation)
                rank = permutation._alphabet.rank
                for letter, length in lengths.items():
                    l[rank(letter)] = length
                lengths = l
            self._lengths = vector(lengths)
            if self._lengths is lengths:
                # NOTE: vector(.) never performs copy
                self._lengths = self._lengths.__copy__()
            V = self._lengths.parent()
            self._base_ring = self._lengths.base_ring()
            self._vector_space = V.ambient_module()

    def vector_space(self):
        return self._vector_space

    def base_ring(self):
        r"""
        Return the base ring over which the lengths are defined

        EXAMPLES::

            sage: from surface_dynamics import *
            sage: p = iet.Permutation('a b', 'b a')

            sage: T = iet.IntervalExchangeTransformation(p, [3, 12])
            sage: T.base_ring()
            Integer Ring

            sage: T = iet.IntervalExchangeTransformation(p, [3, 12/5])
            sage: T.base_ring()
            Rational Field

            sage: T = iet.IntervalExchangeTransformation(p, [3, AA(12).sqrt()])
            sage: T.base_ring()
            Algebraic Real Field

            sage: sqrt2 = QuadraticField(2).gen()
            sage: T = iet.IntervalExchangeTransformation(p, [1, sqrt2])
            sage: T.base_ring()
            Number Field in a ...
        """
        return self._base_ring

    def permutation(self):
        r"""
        Returns the permutation associated to this iet.

        OUTPUT:

        permutation -- the permutation associated to this iet

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: perm = iet.Permutation('a b c','c b a')
            sage: p = iet.IntervalExchangeTransformation(perm,(1,2,1))
            sage: p.permutation() == perm
            True
        """
        return copy(self._permutation)

    def length(self, label=None):
        r"""
        Return the length of a subinterval, or if ``label`` is None the total length.

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: t = iet.IntervalExchangeTransformation(('a b','b a'), [1,3])
            sage: t.length()
            4
            sage: t.length('a')
            1
            sage: t.length('b')
            3
        """
        if label is None:
            return sum(self._lengths)
        else:
            i = self._permutation._alphabet.rank(label)
            return self._lengths[i]

    def lengths(self):
        r"""
        Return the list of lengths associated to this iet.

        OUTPUT:

        vector -- the list of lengths of subinterval (the order of the entries
                  correspond to the alphabet)

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: p = iet.IntervalExchangeTransformation(('a b','b a'),[1,3])
            sage: p.lengths()
            (1, 3)
        """
        return self._lengths.__copy__()

    def translations(self):
        r"""
        Return the vector of translations operated on each intervals.

        EXAMPLES::

            sage: from surface_dynamics import *
            sage: p = iet.Permutation('a b c', 'c b a')
            sage: T = iet.IntervalExchangeTransformation(p, [5,1,3])
            sage: T.translations()
            (4, -2, -6)

        The order of the entries correspond to the alphabet::

            sage: p = iet.Permutation('a c d b', 'b d c a', alphabet='abcd')
            sage: T = iet.IntervalExchangeTransformation(p, [1, 1, 1, 1])
            sage: T.translations()
            (3, -3, 1, -1)

        This vector is covariant with respect to the Rauzy matrices::

            sage: p = iet.Permutation('a b c d', 'd c b a')
            sage: R = p.rauzy_diagram()
            sage: g = R.path(p, *'ttbtbtbtbb')
            sage: T = g.self_similar_iet()
            sage: for i in range(12):
            ....:     S, code = T.zorich_move(iterations=i, data=True)
            ....:     gg = R.path(p, *code)
            ....:     m = gg.matrix()
            ....:     assert m * S.lengths() == T.lengths()
            ....:     assert m.transpose() * T.translations() == S.translations()
        """
        dom_sg = self.domain_singularities()
        im_sg = self.range_singularities()

        p = self._permutation._labels
        top_twin = self._permutation._twin[0]
        top = p[0]
        bot = p[1]

        translations = self.vector_space()()
        for i0,j in enumerate(top):
            i1 = top_twin[i0]
            translations[j] = im_sg[i1] - dom_sg[i0]

        return translations

    def sah_arnoux_fathi_invariant(self):
        r"""
        Return the Sah-Arnoux-Fathi invariant

        The interval exchange needs to be defined over a number field. The output
        is then a vector with rational entries of dimension `d (d-1) / 2` where
        `d` is the degree of the field.

        EXAMPLES:

        The golden rotation::

            sage: from surface_dynamics import *
            sage: p = iet.Permutation('a b','b a')
            sage: R = p.rauzy_diagram()
            sage: g = R.path(p, 't', 'b')
            sage: T = g.self_similar_iet()
            sage: T.sah_arnoux_fathi_invariant()
            (2)

        The Sah-Arnoux-Fathi invariant is not changed under Rauzy
        (or Zorich) induction::

            sage: S = T.zorich_move(iterations=100)
            sage: S.sah_arnoux_fathi_invariant()
            (2)
            sage: (T.length().n(), S.length().n())
            (2.61803398874989, 0.000000000000000)

        An other rotation::

            sage: g = R.path(p, 't', 'b', 'b')
            sage: T = g.self_similar_iet()
            sage: T.sah_arnoux_fathi_invariant()
            (1)
            sage: T.rauzy_move().sah_arnoux_fathi_invariant()
            (1)

        Arnoux-Yoccoz in genus 3::

            sage: x = polygen(ZZ)
            sage: poly = x^3 - x^2 - x - 1
            sage: l = max(poly.roots(AA, False))
            sage: K.<a> = NumberField(poly, embedding=l)
            sage: top = 'A1l A1r A2 B1 B2 C1 C2'
            sage: bot = 'A1r B2 B1 C2 C1 A2 A1l'
            sage: p = iet.Permutation(top, bot)
            sage: lengths = vector((a+1, a**2-a-1, a**2, a, a, 1, 1))
            sage: T = iet.IntervalExchangeTransformation(p, lengths)
            sage: T.sah_arnoux_fathi_invariant()
            (0, 0, 0)

        Arnoux-Yoccoz examples in genus 4 (an auto-similar iet)::

            sage: x = polygen(ZZ)
            sage: poly = x^4 - x^3 - x^2 - x - 1
            sage: l = max(poly.roots(AA, False))
            sage: K.<a> = NumberField(poly, embedding=l)
            sage: top = 'A1l A1r A2 B1 B2 C1 C2 D1 D2'
            sage: bot = 'A1r B2 B1 C2 C1 D2 D1 A2 A1l'
            sage: p = iet.Permutation(top, bot)
            sage: lengths = vector((a**4-a**3, 2*a**3-a**4, a**3, a**2, a**2, a, a, 1, 1))
            sage: T = iet.IntervalExchangeTransformation(p, lengths)
            sage: T.sah_arnoux_fathi_invariant()
            (0, 0, 0, 0, 0, 0)
            sage: T.zorich_move(iterations=10).sah_arnoux_fathi_invariant()
            (0, 0, 0, 0, 0, 0)

        A cubic example in genus 4 (H_4(6)^hyp)::

            sage: p = iet.Permutation([0, 1, 2, 3, 4, 5, 6, 7], [7, 6, 5, 4, 3, 2, 1, 0])
            sage: x = polygen(QQ)
            sage: poly = x^3 - 6*x^2 + 9*x - 3
            sage: emb = AA.polynomial_root(poly, RIF(1.64, 1.66))
            sage: K.<A> = NumberField(poly, embedding=emb)
            sage: lengths= (-4*A^2 + 4*A + 18, -2*A^2 + 19*A - 25, 168*A^2 - 355*A + 128,
            ....:           586*A^2 - 1317*A + 576, -725*A^2 + 1626*A - 707,
            ....:           -11*A^2 - 6*A + 41, -32*A^2 + 100*A - 77, 20*A^2 - 71*A + 63)
            sage: S = iet.IntervalExchangeTransformation(p, lengths)
            sage: S.permutation().stratum_component()
            H_4(6)^hyp
            sage: S.sah_arnoux_fathi_invariant()
            (0, 0, 0)
            sage: S.zorich_move('left', 120).normalize(17) == S
            True

        A quartic example in genus 4 (H_4(6)^even)::

            sage: p = iet.Permutation([0, 1, 2, 3, 4, 5, 6, 7], [7, 1, 0, 5, 4, 3, 2, 6])
            sage: poly = x^4 - 7*x^3 + 14*x^2 - 8*x + 1
            sage: emb = AA.polynomial_root(poly, RIF(0.17, 0.18))
            sage: K.<A> = NumberField(poly, embedding=emb)
            sage: lengths = (2*A^3 - 18*A^2 + 44*A - 4, -10*A^3 + 55*A^2 - 60*A + 10,
            ....:    15*A^3 - 90*A^2 + 120*A - 15, -16*A^3 + 94*A^2 - 132*A + 22,
            ....:    9*A^3 - 56*A^2 + 88*A - 13, 4*A^3 - 16*A^2 + 3*A + 2,
            ....:    -8*A^3 + 57*A^2 - 101*A + 16, 4*A^3 - 26*A^2 + 38*A - 3)
            sage: S = iet.IntervalExchangeTransformation(p, lengths)
            sage: S.permutation().stratum_component()
            H_4(6)^even
            sage: S.sah_arnoux_fathi_invariant()
            (0, 0, 0, 0, 0, 0)
            sage: S.zorich_move('left', 38).normalize(15) == S
            True
        """
        if self.base_ring() is ZZ:
            return free_module_element(ZZ, [])
        elif self.base_ring() is QQ:
            return free_module_element(QQ, [])

        try:
            K, from_V, to_V = self.base_ring().vector_space()
        except (AttributeError, ValueError):
            raise ValueError("the interval exchange needs to be define over a number field")

        return sum(wedge(to_V(u), to_V(v)) for u,v in zip(self.lengths(), self.translations()))

    def normalize(self, total=1, inplace=False):
        r"""
        Returns a interval exchange transformation of normalized lengths.

        The normalization consist in consider a constant homothetic value for
        each lengths in such way that the sum is given (default is 1).

        INPUT:

        - ``total`` - (default: 1) The total length of the interval

        OUTPUT:

        iet -- the normalized iet

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: t = iet.IntervalExchangeTransformation(('a b','b a'), [1,3])
            sage: t.length()
            4
            sage: s = t.normalize(2)
            sage: s.length()
            2
            sage: s.lengths()
            (1/2, 3/2)
        """
        try:
            y = float(total)
        except ValueError:
            raise TypeError("unable to convert x (='%s') into a real number"  %(str(x)))

        if total <= 0:
           raise ValueError("the total length must be positive")

        if inplace:
            res = self
        else:
            res = copy(self)

        res._lengths *= total / res.length()
        if not inplace:
            return res

    def __repr__(self):
        r"""
        A representation string.

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: a = iet.IntervalExchangeTransformation(('a','a'),[1])
            sage: a   #indirect doctest
            Interval exchange transformation of [0, 1[ with permutation
            a
            a
        """
        interval = "[0, %s["%self.length()
        s = "Interval exchange transformation of %s "%interval
        s += "with permutation\n%s"%self._permutation
        return s

    def is_identity(self):
        r"""
        Returns True if self is the identity.

        OUTPUT:

        boolean -- the answer

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: p = iet.Permutation("a b","b a")
            sage: q = iet.Permutation("c d","d c")
            sage: s = iet.IET(p, [1,5])
            sage: t = iet.IET(q, [5,1])
            sage: (s*t).is_identity()
            True
            sage: (t*s).is_identity()
            True
        """
        return self._permutation.is_identity()

    def inverse(self):
        r"""
        Returns the inverse iet.

        OUTPUT:

        iet -- the inverse interval exchange transformation

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: p = iet.Permutation("a b","b a")
            sage: s = iet.IET(p, [1,sqrt(2)-1])
            sage: t = s.inverse()
            sage: t.permutation()
            b a
            a b
            sage: t.lengths()
            (1, sqrt(2) - 1)
            sage: t*s
            Interval exchange transformation of [0, sqrt(2)[ with permutation
            aa bb
            aa bb

        We can verify with the method .is_identity()::

            sage: p = iet.Permutation("a b c d","d a c b")
            sage: s = iet.IET(p, [1, sqrt(2), sqrt(3), sqrt(5)])
            sage: (s * s.inverse()).is_identity()
            True
            sage: (s.inverse() * s).is_identity()
            True
        """
        res = copy(self)
        res._permutation = self._permutation.top_bottom_inverse()
        return res

    def erase_marked_points(self):
        """
        Remove the marked points.

        EXAMPLES::

            sage: from surface_dynamics import *
            sage: p = iet.Permutation([0,1,2,3,4], [4,2,3,0,1])
            sage: T1 = iet.IntervalExchangeTransformation(p, [13,2,4,5,7])
            sage: T1.erase_marked_points()
            Interval exchange transformation of [0, 40[ with permutation
            0 4
            4 0

        TESTS:

            sage: K.<a> = NumberField(x^5 - 2, embedding=AA(2)**(1/5))

        Left side fake zero::

            sage: p = iet.Permutation([1,5,3,4,6,2], [3,2,5,1,4,6])
            sage: T1 = iet.IntervalExchangeTransformation(p, [1, a, a**2, a**3, a**4, 1 + a])
            sage: T2 = T1.erase_marked_points()
            sage: T2
            Interval exchange transformation of [0, a^4 + a^3 + a^2 + 3*a + 2[ with permutation
            1 3 4 2
            3 2 1 4
            sage: T2.permutation().stratum()
            H_2(2)
            sage: assert T2.length(1) == T1.length(1) + T1.length(5)
            sage: assert T2.length(2) == T1.length(2)
            sage: assert T2.length(3) == T1.length(3) + T1.length(5)
            sage: assert T2.length(4) == T1.length(4) + T1.length(6)
            sage: assert T1.sah_arnoux_fathi_invariant() == T2.sah_arnoux_fathi_invariant()

        Right side fake zero::

            sage: p = iet.Permutation([1,3,4,5,2], [3,2,5,1,4])
            sage: T1 = iet.IntervalExchangeTransformation(p, [1,a,a**2,a**3,a**4])
            sage: T2 = T1.erase_marked_points()
            sage: T2
            Interval exchange transformation of [0, a^4 + 2*a^3 + a^2 + a + 1[ with permutation
            1 3 4 2
            3 2 1 4
            sage: T2.permutation().stratum()
            H_2(2)
            sage: assert T2.length(1) == T1.length(1)
            sage: assert T2.length(2) == T1.length(2) + T1.length(5)
            sage: assert T2.length(3) == T1.length(3)
            sage: assert T2.length(4) == T1.length(4) + T1.length(5)
            sage: assert T1.sah_arnoux_fathi_invariant() == T2.sah_arnoux_fathi_invariant()

        Left and right sides fake zeros::

            sage: p = iet.Permutation([1,5,3,4,6,2], [3,2,6,5,1,4])
            sage: T1 = iet.IntervalExchangeTransformation(p, [1,a+1,a**2+1,a**3,a**4-a,a**3+a+1])
            sage: T2 = T1.erase_marked_points()
            sage: T2
            Interval exchange transformation of [0, 2*a^4 + 2*a^3 + a^2 + a + 5[ with permutation
            1 3 4 2
            3 2 1 4
            sage: T2.permutation().stratum()
            H_2(2)
            sage: assert T2.length(1) == T1.length(1) + T1.length(5)
            sage: assert T2.length(2) == T1.length(2) + T1.length(6)
            sage: assert T2.length(3) == T1.length(3) + T1.length(5)
            sage: assert T2.length(4) == T1.length(4) + T1.length(6)
            sage: assert T1.sah_arnoux_fathi_invariant() == T2.sah_arnoux_fathi_invariant()

        Left-right fake zero::

            sage: p = iet.Permutation([1,3,4,2,5], [2,3,5,1,4])
            sage: T1 = iet.IntervalExchangeTransformation(p, [1, a+1, a**2, a**3, a**4-a])
            sage: T2 = T1.erase_marked_points()
            sage: T2
            Interval exchange transformation of [0, 2*a^4 + 2*a^3 + a^2 - a + 2[ with permutation
            1 3 4 2
            2 3 1 4
            sage: T2.permutation().stratum()
            H_2(2)
            sage: assert T2.length(1) == T1.length(1) + T1.length(5)
            sage: assert T2.length(2) == T1.length(2) + T1.length(5)
            sage: assert T2.length(3) == T1.length(3)
            sage: assert T2.length(4) == T1.length(4) + T1.length(2)
            sage: assert T1.sah_arnoux_fathi_invariant() == T2.sah_arnoux_fathi_invariant()

        A small example that end up wrong::

            sage: p = iet.Permutation([0,2,1,3,4], [4,0,1,2,3])
            sage: K.<a> = NumberField(x^2 - 2, embedding=AA(2)**(1/2))
            sage: lengths = [1, a + 1, a, a + 1, 2*a - 2]
            sage: T = iet.IntervalExchangeTransformation(p, lengths)
            sage: U = T.erase_marked_points()
            sage: U.permutation().stratum() 
            H_2(2)
            sage: assert T.sah_arnoux_fathi_invariant() == U.sah_arnoux_fathi_invariant()

            sage: top = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16]
            sage: bot = [15, 11, 12, 9, 10, 7, 8, 5, 6, 2, 3, 4, 16, 14, 0, 1, 13]
            sage: p = iet.Permutation(top, bot)
            sage: x = polygen(QQ)
            sage: poly = x^6 - 6*x^4 + 9*x^2 - 3
            sage: emb = AA.polynomial_root(poly, RIF(1.25, 1.30))
            sage: K.<a> = NumberField(poly, embedding=emb)
            sage: lengths = (-2*a^4 + 7*a^2 - 6, 10/3*a^4 - 17*a^2 + 19, -a^4 + 6*a^2 - 7,
            ....:     -1/3*a^4 + 3*a^2 - 4, -2/3*a^4 + 3*a^2 - 3, 2/3*a^4 + 11*a^2 - 20,
            ....:     2/3*a^4 + 11*a^2 - 20, 8*a^4 - 35*a^2 + 36, 8*a^4 - 35*a^2 + 36,
            ....:     -29/3*a^4 + 42*a^2 - 43, -29/3*a^4 + 42*a^2 - 43, 8/3*a^4 - 28*a^2 + 39,
            ....:     8/3*a^4 - 28*a^2 + 39, -4*a^4 + 20*a^2 - 22, 4/3*a^4 - 4*a^2 + 3,
            ....:     -8/3*a^4 + 16*a^2 - 19, 8/3*a^4 - 14*a^2 + 16)
            sage: T = iet.IntervalExchangeTransformation(p, lengths)
            sage: T.permutation().stratum_component()
            H_4(6, 0^9)^hyp
            sage: T.sah_arnoux_fathi_invariant()
            (0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0)
            sage: U = T.erase_marked_points()
            sage: U.permutation().stratum_component()
            H_4(6)^hyp
            sage: U.sah_arnoux_fathi_invariant()
            (0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0)
        """
        p = self._permutation.__copy__()
        n = len(p)
        lengths = self._lengths
        tt, tb = self._permutation._twin
        lt, lb = self._permutation._labels
        newlengths = list(self._lengths)
        assert max(lt) < len(newlengths)
        assert max(lb) < len(newlengths)
        remove = set()
        newtop = []
        i = 0

        # singularities strictly in the middle
        while i < n:
            k = 1
            while i + k < n and tt[i] + k == tt[i + k]:
                k += 1
            newlengths[lt[i]] = sum(lengths[j] for j in range(i, i+k))
            for j in range(i+1, i+k):
                remove.add(lt[j])
            newtop.append(lt[i])
            i += k

        if remove:
            newbot = [j for j in lb if j not in remove]
            assert set(newtop) == set(newbot)
            p._labels = [newtop, newbot]
            p._init_twin(p._labels)
            tt, tb = p._twin
            lt, lb = newtop, newbot
            n = len(lt)

        i0 = tb[0] - 1
        i1 = tt[0] - 1
        if len(newtop) > 2  and i0 > 0 and i1 > 0 and tt[i0] == i1:
            # left end fake zero
            # the permutation looks like
            # A ... C B ...
            # B ... C A ...
            # backward top-left
            # A ... C B ...   l(A) <- l(A) + l(C)
            # C B ... A ...
            # fusion C in B
            # A ... B ...
            # B ... A ...
            assert lt[i0] == lb[i1]
            A = lt[0]
            B = lb[0]
            C = lt[i0]
            newlengths[A] += newlengths[C]
            newlengths[B] += newlengths[C]
            newlengths[C] = 0
            newtop.pop(newtop.index(C))

            newbot = [i for i in lb if i != C]
            p._labels = [newtop, newbot]
            assert set(newtop) == set(newbot)
            p._init_twin(p._labels)
            tt, tb = p._twin
            lt, lb = newtop, newbot
            n = len(lt)

        i0 = tb[-1] + 1
        i1 = tt[-1] + 1
        if len(newtop) > 2 and i0 < n and i1 < n and tt[i0] == i1:
            # right end fake zero
            # the permutation looks like
            # ... B C ... A
            # ... A C ... B
            # backward top-right
            # ... B C ... A
            # ... A ... B C
            # fusion C in B
            # ... B ... A
            # ... A ... B
            assert lt[i0] == lb[i1]
            A = lt[-1]
            B = lb[-1]
            C = lt[i0]
            assert newtop[-1] == A
            newlengths[A] += newlengths[C]
            newlengths[B] += newlengths[C]
            newlengths[C] = 0
            newtop.pop(newtop.index(C))

            newbot = [i for i in lb if i != C]
            assert set(newtop) == set(newbot)
            p._labels = [newtop, newbot]
            p._init_twin(p._labels)
            tt, tb = p._twin
            lt, lb = newtop, newbot
            n = len(lt)

        i0 = tb[0] - 1 # D
        i1 = tt[0] - 1 # C
        if len(newtop) > 2 and tb[i1] == tt[i0] == n-1:
            # left-right end fake zero
            # the permutation looks like
            # A ... D B ... C  or  A ... D B  or  A B ... C
            # B ... C A ... D      B A ... D      B ... C A
            #                      (B=C case)     (A=D case)
            assert lb[-1] == lt[i0] and lt[-1] == lb[i1]
            A = lt[0]
            B = lb[0]
            C = lb[i1]
            D = lt[i0]
            assert newtop[0] == A and newtop[-1] == C and B != D
            assert A in newtop and B in newtop and C in newtop and D in newtop
            if B == C:
                # A ... D B
                # B A ... D
                # backward bot-left
                # D A ... B    l(B) <- l(B) + l(D)
                # B A ... D
                # backward top-right
                # D A ... B    l(B) <- l(B) + l(A)
                # B ... D A
                # fusion A in D
                # D ... B
                # B ... D
                assert newtop[0] == A and newtop[-1] == B and newtop[-2] == D
                newlengths[B] += newlengths[A] + newlengths[D]
                newlengths[D] += newlengths[A]
                newlengths[A] = 0
                newtop.pop(0)
                newtop.pop(-2)
                newtop.insert(0, D)
                assert newtop[0] == D and newtop[-1] == B

                newbot = [i for i in lb if i != A]

            elif A == D:
                # A B ... C
                # B ... C A
                # backward bot-right
                # A ... C B   l(A) <- l(A) + l(B)
                # B ... C A
                # backward top-left
                # A ... C B   l(A) <- l(A) + l(C)
                # C B ... A
                # fusion C in B
                # A ... B
                # B ... A
                newlengths[A] += newlengths[B] + newlengths[C]
                newlengths[B] += newlengths[C]
                newlengths[C] = 0
                assert newtop[-1] == C
                assert newtop[1] == B
                newtop.pop(-1)
                newtop.pop(1)
                newtop.append(B)
                assert newtop[0] == A and newtop[-1] == B, (newtop, A, B)
                assert C not in newtop
                newbot = [i for i in lb if i != C]

            else:
                # do backward Rauzy top-left / bottom-right
                # A ... D ... C B     l(A) <- l(A) + l(C)
                # C B ... A ... D     l(D) <- l(D) + l(B)
                # fusion C in B
                # A ... D ... B
                # B ... A ... D
                newlengths[A] += newlengths[C]  # backward rauzy
                newlengths[D] += newlengths[B]  # backward rauzy
                newlengths[B] += newlengths[C]  # absorbtion
                newlengths[C] = 0
                newtop.pop(newtop.index(C))
                newtop.pop(newtop.index(B))
                newtop.append(B)
                newbot = [i for i in lb if i != C]

            assert set(newtop) == set(newbot)
            p._labels = [newtop, newbot]
            p._init_twin(p._labels)
            tt, tb = p._twin
            lt, lb = newtop, newbot

        unrank = self._permutation.alphabet().unrank
        newtop = [unrank(lab) for lab in lt]
        newbot = [unrank(lab) for lab in lb]
        newlengths = [newlengths[j] for j in lt]
        p = LabelledPermutationIET([newtop, newbot])
        return IntervalExchangeTransformation(p, newlengths)

    def __mul__(self, other):
        r"""
        Composition of iet.

        The domain (i.e. the length) of the two iets must be the same). The
        alphabet choosen depends on the permutation.

        TESTS:

            sage: from surface_dynamics import *

        ::

            sage: p = iet.Permutation("a b", "a b")
            sage: t = iet.IET(p, [1,1])
            sage: r = t*t
            sage: r.permutation()
            aa bb
            aa bb
            sage: r.lengths()
            (1, 1)

        ::

            sage: p = iet.Permutation("a b","b a")
            sage: t = iet.IET(p, [1,1])
            sage: r = t*t
            sage: r.permutation()
            ab ba
            ab ba
            sage: r.lengths()
            (1, 1)

        ::

            sage: p = iet.Permutation("1 2 3 4 5","5 4 3 2 1")
            sage: q = iet.Permutation("a b","b a")
            sage: s = iet.IET(p, [1]*5)
            sage: t = iet.IET(q, [1/2, 9/2])
            sage: r = s*t
            sage: r.permutation()
            a5 b1 b2 b3 b4 b5
            b5 a5 b4 b3 b2 b1
            sage: r.lengths()
            (1/2, 1, 1, 1, 1, 1/2)
            sage: r = t*s
            sage: r.permutation()
            1b 2b 3b 4b 5a 5b
            5b 4b 3b 2b 1b 5a
            sage: r.lengths()
            (1, 1, 1, 1, 1/2, 1/2)
            sage: t = iet.IET(q, [3/2, 7/2])
            sage: r = s*t
            sage: r.permutation()
            a4 a5 b1 b2 b3 b4
            a5 b4 a4 b3 b2 b1
            sage: r.lengths()
            (1/2, 1, 1, 1, 1, 1/2)
            sage: t = iet.IET(q, [5/2,5/2])
            sage: r = s*t
            sage: r.permutation()
            a3 a4 a5 b1 b2 b3
            a5 a4 b3 a3 b2 b1
            sage: r = t*s
            sage: r.permutation()
            1b 2b 3a 3b 4a 5a
            3b 2b 1b 5a 4a 3a

        ::

            sage: p = iet.Permutation("a b","b a")
            sage: s = iet.IET(p, [4,2])
            sage: q = iet.Permutation("c d","d c")
            sage: t = iet.IET(q, [3, 3])
            sage: r1 = t * s
            sage: r1.permutation()
            ac ad bc
            ad bc ac
            sage: r1.lengths()
            (1, 3, 2)
            sage: r2 = s * t
            sage: r2.permutation()
            ca cb da
            cb da ca
            sage: r2.lengths()
            (1, 2, 3)
        """
        assert(
            isinstance(other, IntervalExchangeTransformation) and
            self.length() == other.length())

        from .labelled import LabelledPermutationIET
        from sage.combinat.words.words import Words

        other_sg = other.range_singularities()[1:]
        self_sg = self.domain_singularities()[1:]

        n_other = len(other._permutation)
        n_self = len(self._permutation)

        interval_other = other._permutation._labels[1]
        interval_self = self._permutation._labels[0]

        d_other = dict([(i,[]) for i in interval_other])
        d_self = dict([(i,[]) for i in interval_self])

        i_other = 0
        i_self = 0

        x = self.base_ring().zero()
        l_lengths = []
        while i_other < n_other and i_self < n_self:
            j_other = interval_other[i_other]
            j_self = interval_self[i_self]

            d_other[j_other].append(j_self)
            d_self[j_self].append(j_other)

            if other_sg[i_other] < self_sg[i_self]:
                l = other_sg[i_other] - x
                x = other_sg[i_other]
                i_other += 1
            elif other_sg[i_other] > self_sg[i_self]:
                l = self_sg[i_self] - x
                x = self_sg[i_self]
                i_self += 1
            else:
                l = self_sg[i_self] - x
                x = self_sg[i_self]
                i_other += 1
                i_self += 1

            l_lengths.append(((j_other,j_self),l))

        alphabet_other = other._permutation.alphabet()
        alphabet_self = self._permutation.alphabet()

        d_lengths = dict(l_lengths)

        l_lengths = []
        top_interval = []
        for i in other._permutation._labels[0]:
            for j in d_other[i]:
                a = alphabet_other.unrank(i)
                b = alphabet_self.unrank(j)
                top_interval.append(str(a)+str(b))
                l_lengths.append(d_lengths[(i,j)])

        bottom_interval = []
        for i in self._permutation._labels[1]:
            for j in d_self[i]:
                a = alphabet_other.unrank(j)
                b = alphabet_self.unrank(i)
                bottom_interval.append(str(a)+str(b))

        p = LabelledPermutationIET((top_interval,bottom_interval))
        return IntervalExchangeTransformation(p,l_lengths)

    def __eq__(self, other):
        r"""
        Tests equality

        TESTS::

            sage: from surface_dynamics import *

            sage: t = iet.IntervalExchangeTransformation(('a b','b a'),[1,1])
            sage: t == t
            True
        """
        return (
            isinstance(self, type(other)) and
            self._permutation == other._permutation and
            self._lengths == other._lengths)

    def __ne__(self, other):
        r"""
        Tests difference

        TESTS::

            sage: from surface_dynamics import *

            sage: t = iet.IntervalExchangeTransformation(('a b','b a'),[1,1])
            sage: t != t
            False
        """
        return (
            not isinstance(self, type(other)) or
            self._permutation != other._permutation or
            self._lengths != other._lengths)

    def recoding(self, n):
        r"""
        Recode this interval exchange transformation on the words of length
        ``n``.

        EXAMPLES::

            sage: from surface_dynamics import *
            sage: p = iet.Permutation('a d c b', 'b c a d', alphabet='abcd')
            sage: T = iet.IntervalExchangeTransformation(p, [119,213,82,33])
            sage: T.recoding(2)
            Interval exchange transformation of [0, 447[ with permutation
            ab db cc cb ba bd bc
            ba bd bc cc cb ab db
            sage: T.recoding(3)
            Interval exchange transformation of [0, 447[ with permutation
            aba abd abc dbc ccb cba bab bdb bcc bcb
            cba aba abd abc dbc bcc bcb ccb bab bdb
        """
        length = len(self._permutation)
        lengths = self._lengths
        A = self.permutation().alphabet()
        unrank = A.unrank

        top = self._permutation._labels[0]
        bot = self._permutation._labels[1]
        bot_twin = self._permutation._twin[1]

        sg_top = self.domain_singularities()[1:-1]  # iterates of the top singularities
        cuts = [[[i], lengths[i]] for i in bot]     # the refined bottom interval

        translations = self.translations()

        for step in range(n-1):
            i = 0
            y = self.base_ring().zero()
            new_sg_top = []
            limits = [0]
            for j,x in enumerate(sg_top):
                while y < x:
                    cuts[i][0].append(top[j])
                    y += cuts[i][1]
                    i += 1

                limits.append(i)
                if y != x:
                    cuts.insert(i, [cuts[i-1][0][:-1], cuts[i-1][1]])
                    cuts[i-1][1] -= y-x
                    cuts[i][0].append(top[j+1])
                    cuts[i][1] = y-x
                    i += 1
            while i < len(cuts):
                cuts[i][0].append(top[j+1])
                i += 1
            limits.append(len(cuts))

            # now we reorder cuts with respect to T according to the cut at
            # limits
            if step != n-2:
                new_cuts = []
                for j in bot_twin:
                    new_cuts.extend(cuts[limits[j]:limits[j+1]])
                cuts = new_cuts

        # build the new interval exchange transformations
        itop = [None]*(max(top)+1)
        for i,j in enumerate(top):
            itop[j] = i
        key = lambda x: [itop[i] for i in x[0]]
        top = sorted(cuts, key=key)
        lengths = [y for x,y in top]
        top = [''.join(str(unrank(i)) for i in x) for x,y in top]
        bot = cuts
        bot = [''.join(str(unrank(i)) for i in x) for x,y in bot]

        from .labelled import LabelledPermutationIET
        p = LabelledPermutationIET((top,bot))
        return IntervalExchangeTransformation(p,lengths)

    def in_which_interval(self, x, interval=0):
        r"""
        Returns the letter for which x is in this interval.

        INPUT:

        - ``x`` - a positive number

        - ``interval`` - (default: 'top') 'top' or 'bottom'


        OUTPUT:

        label -- a label corresponding to an interval

        TESTS::

            sage: from surface_dynamics import *

            sage: t = iet.IntervalExchangeTransformation(('a b c','c b a'),[1,1,1])
            sage: t.in_which_interval(0)
            'a'
            sage: t.in_which_interval(0.3)
            'a'
            sage: t.in_which_interval(1)
            'b'
            sage: t.in_which_interval(1.9)
            'b'
            sage: t.in_which_interval(2)
            'c'
            sage: t.in_which_interval(2.1)
            'c'
            sage: t.in_which_interval(3)
            Traceback (most recent call last):
            ...
            ValueError: your value does not lie in [0; 3[

        .. and for the bottom interval::

            sage: t.in_which_interval(0,'bottom')
            'c'
            sage: t.in_which_interval(1.2,'bottom')
            'b'
            sage: t.in_which_interval(2.9,'bottom')
            'a'
        """
        interval = interval_conversion(interval)

        if x < 0 or x >= self.length():
            raise ValueError("your value does not lie in [0; {}[".format(self.length()))

        i = 0

        while x >= 0:
            x -= self._lengths[self._permutation._labels[interval][i]]
            i += 1

        i -= 1
        x += self._lengths[self._permutation._labels[interval][i]]

        j = self._permutation._labels[interval][i]
        return self._permutation._alphabet.unrank(j)

    def singularities(self):
        r"""
        The list of singularities of 'T' and 'T^{-1}'.

        OUTPUT:

        list -- two lists of positive numbers which corresponds to extremities
            of subintervals

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: t = iet.IntervalExchangeTransformation(('a b','b a'),[1/2,3/2])
            sage: t.singularities()
            [[0, 1/2, 2], [0, 3/2, 2]]
        """
        return [self.domain_singularities(), self.range_singularities()]

    def domain_singularities(self):
        r"""
        Returns the list of singularities of T

        OUTPUT:

        list -- positive reals that corresponds to singularities in the top
            interval

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: t = iet.IET(("a b","b a"), [1, sqrt(2)])
            sage: t.domain_singularities()
            [0, 1, sqrt(2) + 1]
        """
        l = [self.base_ring().zero()]
        for j in self._permutation._labels[0]:
            l.append(l[-1] + self._lengths[j])
        return l

    def range_singularities(self):
        r"""
        Returns the list of singularities of `T^{-1}`

        OUTPUT:

        list -- real numbers that are singular for `T^{-1}`


        EXAMPLES::

            sage: from surface_dynamics import *
            sage: t = iet.IET(("a b","b a"), [1, sqrt(2)])
            sage: t.range_singularities()
            [0, sqrt(2), sqrt(2) + 1]
        """
        l = [self.base_ring().zero()]
        for j in self._permutation._labels[1]:
            l.append(l[-1] + self._lengths[j])
        return l

    def __call__(self, value):
        r"""
        Return the image of value by this transformation

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: t = iet.IntervalExchangeTransformation(('a b','b a'),[1/2,3/2])
            sage: t(0)
            3/2
            sage: t(1/2)
            0
            sage: t(1)
            1/2
            sage: t(3/2)
            1
        """
        assert(value >= 0 and value < self.length())

        dom_sg = self.domain_singularities()
        im_sg = self.range_singularities()

        a = self.in_which_interval(value)

        i0 = self._permutation[0].index(a)
        i1 = self._permutation[1].index(a)

        return value - dom_sg[i0] + im_sg[i1]

    def rauzy_move(self, side='right', iterations=1, data=False, error_on_saddles=True):
        r"""
        Performs a Rauzy move.

        INPUT:

        - ``side`` - 'left' (or 'l' or 0) or 'right' (or 'r' or 1)

        - ``iterations`` - integer (default :1) the number of iteration of Rauzy
           moves to perform

        - ``data`` - whether to return also the paths and composition of towers

        - ``error_on_saddles`` - (default: ``True``) whether to stop when a saddle
          is encountered

        OUTPUT:

        - ``iet`` -- the Rauzy move of self

        - ``path`` -- (if ``data=True``) a list of 't' and 'b'

        - ``towers`` -- (if ``data=True``) the towers of the Rauzy induction as a word morphism

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: phi = QQbar((sqrt(5)-1)/2)
            sage: t1 = iet.IntervalExchangeTransformation(('a b','b a'),[1,phi])
            sage: t2 = t1.rauzy_move().normalize(t1.length())
            sage: l2 = t2.lengths()
            sage: l1 = t1.lengths()
            sage: l2[0] == l1[1] and l2[1] == l1[0]
            True

            sage: tt,path,sub = t1.rauzy_move(iterations=3, data=True)
            sage: tt
            Interval exchange transformation of [0, 0.3819660112501051?[ with
            permutation
            a b
            b a
            sage: path
            ['b', 't', 'b']
            sage: sub
            WordMorphism: a->aab, b->aabab

        The substitution can also be recovered from the Rauzy diagram::

            sage: p = t1.permutation()
            sage: p.rauzy_diagram().path(p, *path).substitution() == sub
            True

        An other examples involving 3 intervals::

            sage: t = iet.IntervalExchangeTransformation(('a b c','c b a'),[1,1,3])
            sage: t
            Interval exchange transformation of [0, 5[ with permutation
            a b c
            c b a
            sage: t1 = t.rauzy_move()
            sage: t1
            Interval exchange transformation of [0, 4[ with permutation
            a b c
            c a b
            sage: t2 = t1.rauzy_move()
            sage: t2
            Interval exchange transformation of [0, 3[ with permutation
            a b c
            c b a
            sage: t2.rauzy_move()
            Traceback (most recent call last):
            ...
            ValueError: saddle connection found
            sage: t2.rauzy_move(error_on_saddles=False)
            Interval exchange transformation of [0, 2[ with permutation
            a b
            a b

        Degenerate cases::

            sage: p = iet.Permutation('a b', 'b a')
            sage: T = iet.IntervalExchangeTransformation(p, [1,1])
            sage: T.rauzy_move(error_on_saddles=False)
            Interval exchange transformation of [0, 1[ with permutation
            a
            a
        """
        if data:
            towers = {a:[a] for a in self._permutation.letters()}
            path = []

        side = side_conversion(side)

        res = copy(self)
        for i in range(iterations):
            winner,(a,b,c) = res._rauzy_move(side, error_on_saddles=error_on_saddles)
            if data:
                if winner is None:
                    raise ValueError("does not handle degenerate situations")
                towers[a] = towers[b] + towers[c]
                path.append(winner)

        if data:
            from sage.combinat.words.morphism import WordMorphism
            return res, path, WordMorphism(towers)
        else:
            return res

    def backward_rauzy_move(self, winner, side='right'):
        r"""
        Return a new interval exchange transformation obtained by performing a backward rauzy move.

        EXAMPLES::

            sage: from surface_dynamics import iet
            sage: perm = iet.Permutation('a b c d e f', 'f c b e d a')
            sage: x = polygen(QQ)
            sage: poly = x^3 - x^2 - x - 1
            sage: root = AA.polynomial_root(poly, RIF(1.8, 1.9))
            sage: K.<a> = NumberField(poly, embedding=root)
            sage: T = iet.IntervalExchangeTransformation(perm, [a**2, a-1, a + 2, 3, 2*a - 3, 1])
            sage: T
            Interval exchange transformation of [0, a^2 + 4*a + 2[ with permutation
            a b c d e f
            f c b e d a
            sage: S = T.backward_rauzy_move(0).backward_rauzy_move(1).backward_rauzy_move(1)
            sage: S
            Interval exchange transformation of [0, a^2 + 7*a + 4[ with permutation
            a b c f d e
            f b e d a c
            sage: S.rauzy_move(iterations=3) == T
            True
        """
        winner = interval_conversion(winner)
        side = side_conversion(side)
        res = copy(self)
        res._backward_rauzy_move(winner, side)
        return res

    def zorich_move(self, side='right', iterations=1, data=False):
        r"""
        Performs a Rauzy move.

        INPUT:

        - ``side`` - 'left' (or 'l' or 0) or 'right' (or 'r' or 1)

        - ``iterations`` - integer (default :1) the number of iteration of Rauzy
           moves to perform

        - ``data`` - whether to return also the path

        OUTPUT:

        - ``iet`` -- the Rauzy move of self

        - ``path`` -- (if ``data=True``) a list of 't' and 'b'

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: p = iet.Permutation('a b c', 'c b a')
            sage: T = iet.IntervalExchangeTransformation(p, [12, 35, 67])
            sage: T.zorich_move()
            Interval exchange transformation of [0, 55[ with permutation
            a b c
            c a b
            sage: assert T.permutation() == p and T.lengths() == vector((12,35,67))

        A self similar example in genus 2::

            sage: p = iet.Permutation('a b c d', 'd a c b')
            sage: R = p.rauzy_diagram()
            sage: code = 'b'*4 + 't'*1 + 'b'*3 + 't'*1 + 'b'*3 + 't'*1 + 'b'*1 + 't'*1 + 'b'*4 + 't'*1 + 'b'*2 + 't'*7
            sage: g = R.path(p, *code)
            sage: m = g.matrix()
            sage: poly = m.charpoly()
            sage: l = max(poly.roots(AA, False))
            sage: K.<a> = NumberField(poly, embedding=l)
            sage: lengths = (m - a).right_kernel().basis()[0]
            sage: T = iet.IntervalExchangeTransformation(p, lengths)
            sage: T.normalize(a, inplace=True)
            sage: T
            Interval exchange transformation of [0, a[ with permutation
            a b c d
            d a c b
            sage: T2, path = T.zorich_move(iterations=12, data=True)
            sage: a*T2.lengths() == T.lengths()
            True
            sage: path == code
            True

        Saddle connection detection::

            sage: p = iet.Permutation('a b c', 'c b a')
            sage: T = iet.IntervalExchangeTransformation(p, [41, 22, 135])
            sage: T.zorich_move(iterations=100)
            Traceback (most recent call last):
            ...
            ValueError: saddle connection found
            sage: p = iet.Permutation('a b c d e f', 'f c b e d a')
            sage: T = iet.IntervalExchangeTransformation(p, [41, 132, 22, 135, 55, 333])
            sage: T.zorich_move(iterations=100)
            Traceback (most recent call last):
            ...
            ValueError: saddle connection found
        """
        if data:
            path = ''

        side = side_conversion(side)

        res = copy(self)
        for i in range(iterations):
            winner, m = res._zorich_move(side)
            if data:
                path += winner * m

        if data:
            return res, path
        else:
            return res


    def _rauzy_move(self, side=-1, error_on_saddles=True):
        r"""
        Perform a Rauzy move inplace

        INPUT:

        - ``side`` - must be 0 or -1 (no verification)

        - ``error_on_saddles`` - (default ``True``) whether an error is raised
          when a saddle connections is encountered

        OUTPUT:

        - ``T`` - the iet after Rauzy induction

        - ``winner`` - either ``'t'`` or ``'b'`` (and possibly ``None`` if
          ``error_on_saddles`` is ``False``)

        - ``(a,b,c)`` - a triple of letter such that the towers are obtained by
          applying the substitution `a \mapsto bc`.

        TEST::

            sage: from surface_dynamics import *

            sage: p = iet.Permutation('a b c d', 'd c b a')
            sage: K.<sqrt2> = QuadraticField(2)
            sage: T = iet.IntervalExchangeTransformation(p, [1,1,1,sqrt2])
            sage: T._rauzy_move()
            ('t', ('a', 'a', 'd'))
            sage: T._rauzy_move()
            ('b', ('d', 'b', 'd'))
            sage: T._rauzy_move()
            ('t', ('b', 'b', 'c'))
            sage: T._rauzy_move()
            ('b', ('c', 'b', 'c'))
            sage: T._rauzy_move()
            ('t', ('b', 'b', 'd'))
            sage: T._rauzy_move()
            ('b', ('d', 'c', 'd'))
            sage: T._rauzy_move()
            ('t', ('c', 'c', 'd'))
            sage: T._rauzy_move()
            ('b', ('d', 'a', 'd'))
            sage: T
            Interval exchange transformation of [0, -4*sqrt2 + 7[ with permutation
            a d b c
            d c b a

        Saddle connections::

            sage: p = iet.Permutation('a b c', 'c b a')
            sage: T = iet.IntervalExchangeTransformation(p, [1,2,1])
            sage: T._rauzy_move()
            Traceback (most recent call last):
            ...
            ValueError: saddle connection found
            sage: T._rauzy_move(error_on_saddles=False)
            (None, ('a', 'a', 'c'))
            sage: T.lengths()
            (1, 2, 0)
            sage: T.permutation()
            a b
            a b

            sage: p = iet.Permutation([0, 1, 2, 3], [3, 1, 2, 0])
            sage: T = iet.IntervalExchangeTransformation(p, [13, 2, 22, 13])
            sage: T._rauzy_move(0, error_on_saddles=False)
            (None, (3, 3, 0))
            sage: T
            Interval exchange transformation of [0, 37[ with permutation
            1 2 3
            1 2 3
            sage: print(T._lengths)
            (0, 2, 22, 13)
            sage: print(T._permutation._labels)
            [[1, 2, 3], [1, 2, 3]]
        """
        top = self._permutation._labels[0][side]
        bot = self._permutation._labels[1][side]

        unrank = self._permutation.alphabet().unrank
        top_letter = unrank(top)
        bot_letter = unrank(bot)

        length_top = self._lengths[top]
        length_bot = self._lengths[bot]

        if length_top > length_bot:
            winner = 0 # TODO: this value is ignored
            winner_interval = top
            loser_interval = bot
            abc = (bot_letter, bot_letter, top_letter)
            winner = 't'
        elif length_top < length_bot:
            winner = 1 # TODO: this value is ignored
            winner_interval = bot
            loser_interval = top
            abc = (top_letter, bot_letter, top_letter)
            winner = 'b'
        elif error_on_saddles:
            raise ValueError("saddle connection found")
        else:
            # we do a pseudo-top Rauzy induction and remove the bot interval
            # (the iet get one interval less)
            p = self._permutation
            top = p._labels[0][side]
            bot = p._labels[1][side]
            p._identify_intervals(side)
            self._lengths[top] = 0
            return None, (bot_letter,bot_letter,top_letter)

        self._permutation = self._permutation.rauzy_move(winner=winner, side=side, inplace=True)
        self._lengths[winner_interval] -= self._lengths[loser_interval]

        return winner, abc

    def _backward_rauzy_move(self, winner, side=-1):
        r"""
        Inplace backward rauzy move.

        EXAMPLES::

            sage: from surface_dynamics import iet
            sage: x = polygen(QQ)
            sage: K.<cbrt2> = NumberField(x^3 - 2, embedding=AA(2)**(1/3))
            sage: p = iet.Permutation("a b c d", "d c b a")
            sage: T0 = iet.IntervalExchangeTransformation(p, [1, cbrt2, cbrt2**2, cbrt2 - 1])
            sage: T1 = copy(T0)
            sage: T1.lengths()
            (1, cbrt2, cbrt2^2, cbrt2 - 1)
            sage: T1._backward_rauzy_move(0)
            sage: T1.lengths()
            (1, cbrt2, cbrt2^2, cbrt2^2 + cbrt2 - 1)
            sage: _ = T1._rauzy_move()
            sage: T1 == T0
            True

            sage: T1._backward_rauzy_move(0)
            sage: T1._backward_rauzy_move(1)
            sage: T1._backward_rauzy_move(1)
            sage: T1._backward_rauzy_move(0)
            sage: _ = T1._rauzy_move()
            sage: _ = T1._rauzy_move()
            sage: _ = T1._rauzy_move()
            sage: _ = T1._rauzy_move()
            sage: T1 == T0
            True
        """
        self._permutation = self._permutation.backward_rauzy_move(winner=winner, side=side, inplace=True)
        winner_interval = self._permutation._labels[winner][side]
        loser_interval = self._permutation._labels[1-winner][side]
        self._lengths[winner_interval] += self._lengths[loser_interval]

    def _zorich_move(self, side=-1):
        r"""
        Performs a Zorich move (acceleration of Rauzy) inplace

        INPUT:

        - ``side`` - must be 0 or -1 (no verification)

        OUTPUT:

        - ``winner_letter`` - whether top or bot was done

        - ``m`` -- number of Rauzy made on the same side

        TEST::

            sage: from surface_dynamics import *

            sage: p = iet.Permutation('a b', 'b a')
            sage: K.<sqrt3> = QuadraticField(3)
            sage: t = iet.IntervalExchangeTransformation(p, [1, sqrt3 - 1])
            sage: t._zorich_move()
            ('b', 1)
            sage: t._zorich_move()
            ('t', 2)
            sage: t._zorich_move()
            ('b', 1)
            sage: t._zorich_move()
            ('t', 2)
            sage: t._zorich_move()
            ('b', 1)
            sage: continued_fraction(sqrt3 - 1)
            [0; (1, 2)*]

            sage: x = polygen(ZZ)
            sage: K.<a> = NumberField(x^3 - 2, embedding=AA(2)**(1/3))
            sage: t = iet.IntervalExchangeTransformation(p, [1, a])
            sage: [t._zorich_move()[1] for _ in range(10)]
            [1, 3, 1, 5, 1, 1, 4, 1, 1, 8]
            sage: continued_fraction(a)
            [1; 3, 1, 5, 1, 1, 4, 1, 1, 8, 1, 14, 1, 10, 2, 1, 4, 12, 2, 3, ...]

        A self-similar example in genus 2::

            sage: x = polygen(ZZ)
            sage: poly = x^4 - 11*x^3 + 21*x^2 - 11*x + 1
            sage: l = max(poly.roots(AA, False))
            sage: K.<a> = NumberField(poly, embedding=l)
            sage: lengths = vector([-3*a^3 + 33*a^2 - 63*a + 33, -4*a^3 + 42*a^2 - 63*a + 14,
            ....:     7*a^3 - 74*a^2 + 116*a - 34, -a^2 + 10*a - 10])
            sage: p = iet.Permutation('a b c d', 'd c b a')
            sage: t = iet.IntervalExchangeTransformation(p, lengths)
            sage: [t._zorich_move() for _ in range(6)]
            [('t', 1), ('b', 2),  ('t', 2), ('b', 1), ('t', 2), ('b', 2)]
            sage: lengths == a * t.lengths()
            True
            sage: [t._zorich_move(side=0) for _ in range(6)]
            [('b', 1), ('t', 2), ('b', 2), ('t', 1), ('b', 2), ('t', 2)]
            sage: lengths == a^2 * t.lengths()
            True

        Saddle connection detection::

            sage: p = iet.Permutation('a b c', 'c b a')
            sage: T = iet.IntervalExchangeTransformation(p, [41, 1, 5])
            sage: T._zorich_move()
            Traceback (most recent call last):
            ...
            ValueError: saddle connection found
        """
        top = self._permutation._labels[0][side]
        bot = self._permutation._labels[1][side]

        length_top = self._lengths[top]
        length_bot = self._lengths[bot]

        if length_top > length_bot:
            winner = 0
            winner_letter = 't'
            winner_interval = top
            loser_interval = bot
        elif length_top < length_bot:
            winner = 1
            winner_letter = 'b'
            winner_interval = bot
            loser_interval = top
        else:
            raise ValueError("saddle connection found")

        # number of full loops
        lwin = self._lengths[winner_interval]
        loser = 1 - winner
        loser_row = self._permutation._labels[loser]
        llos = 0
        k = 0
        if side == -1:
            # right induction
            j = -1
            while loser_row[j] != winner_interval:
                llos += self._lengths[loser_row[j]]
                j -= 1
                k += 1
        else:
            # left induction
            j = 0
            while loser_row[j] != winner_interval:
                llos += self._lengths[loser_row[j]]
                j += 1
                k += 1

        # remove the full loops
        m = (lwin / llos).floor()
        self._lengths[winner_interval] -= m*llos

        # remaining steps
        r = 0
        while self._lengths[winner_interval] >= self._lengths[loser_interval]:
            self._lengths[winner_interval] -= self._lengths[loser_interval]
            self._permutation.rauzy_move(winner=winner, side=side, inplace=True)
            loser_interval = self._permutation._labels[loser][side]
            r += 1

        if self._lengths[winner_interval].is_zero():
            raise ValueError("saddle connection found")

        return winner_letter, k*m + r

    def __copy__(self):
        r"""
        Returns a copy of this interval exchange transformation.

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: t = iet.IntervalExchangeTransformation(('a b','b a'),[1,1])
            sage: s = copy(t)
            sage: s == t
            True
            sage: s is t
            False
        """
        res = self.__class__()
        res._base_ring = self._base_ring
        res._vector_space = self._vector_space
        res._permutation = copy(self._permutation)
        res._lengths = copy(self._lengths)
        return res

    def plot_function(self,**d):
        r"""
        Return a plot of the interval exchange transformation as a
        function.

        INPUT:

        - Any option that is accepted by line2d

        OUTPUT:

        2d plot -- a plot of the iet as a function

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: t = iet.IntervalExchangeTransformation(('a b c d','d a c b'),[1,1,1,1])
            sage: t.plot_function(rgbcolor=(0,1,0))    # not tested (problem with matplotlib font cache)
            Graphics object consisting of 4 graphics primitives
        """
        from sage.plot.plot import Graphics
        from sage.plot.line import line2d

        G = Graphics()
        l = self.singularities()
        t = self._permutation._twin

        for i in range(len(self._permutation)):
            j = t[0][i]
            G += line2d([(l[0][i],l[1][j]),(l[0][i+1],l[1][j+1])],**d)

        return G

    def plot_two_intervals(
        self,
        position=(0,0),
        vertical_alignment = 'center',
        horizontal_alignment = 'left',
        interval_height=0.1,
        labels_height=0.05,
        fontsize=14,
        labels=True,
        colors=None):
        r"""
        Returns a picture of the interval exchange transformation.

        INPUT:

        - ``position`` - a 2-uple of the position

        - ``horizontal_alignment`` - left (defaut), center or right

        - ``labels`` - boolean (defaut: True)

        - ``fontsize`` - the size of the label


        OUTPUT:

        2d plot -- a plot of the two intervals (domain and range)

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: t = iet.IntervalExchangeTransformation(('a b','b a'),[1,1])
            sage: t.plot_two_intervals()
            Graphics object consisting of 8 graphics primitives
        """
        from sage.plot.plot import Graphics
        from sage.plot.line import line2d
        try:
            from sage.plot.plot import text
        except ImportError:
            from sage.plot.text import text
        from sage.plot.colors import rainbow

        G = Graphics()

        lengths = list(map(float, self._lengths))
        total_length = sum(lengths)

        if colors is None:
            colors = rainbow(len(self._permutation), 'rgbtuple')

        if horizontal_alignment == 'left':
            s = position[0]
        elif horizontal_alignment == 'center':
            s = position[0] - total_length / 2
        elif horizontal_alignment == 'right':
            s = position[0] - total_length
        else:
            raise ValueError("horizontal_alignement must be left, center or right")

        top_height = position[1] + interval_height
        for i in self._permutation._labels[0]:
            G += line2d([(s,top_height),(s+lengths[i],top_height)],
                rgbcolor=colors[i])
            if labels == True:
                G += text(str(self._permutation._alphabet.unrank(i)),
                    (s+float(lengths[i])/2,top_height+labels_height),
                    horizontal_alignment='center',
                    rgbcolor=colors[i],
                    fontsize=fontsize)

            s += lengths[i]

        if horizontal_alignment == 'left':
            s = position[0]
        elif horizontal_alignment == 'center':
            s = position[0] - total_length / 2
        elif horizontal_alignment == 'right':
            s = position[0] - total_length
        else:
            raise ValueError("horizontal_alignement must be left, center or right")

        bottom_height = position[1] - interval_height
        for i in self._permutation._labels[1]:
            G += line2d([(s,bottom_height), (s+lengths[i],bottom_height)],
                rgbcolor=colors[i])
            if labels == True:
                G += text(str(self._permutation._alphabet.unrank(i)),
                    (s+float(lengths[i])/2,bottom_height-labels_height),
                    horizontal_alignment='center',
                    rgbcolor=colors[i],
                    fontsize=fontsize)
            s += lengths[i]

        return G

    def plot_towers(self, iterations, position=(0,0), colors=None):
        """
        Plot the towers of this interval exchange obtained from Rauzy induction.

        INPUT:

        - ``nb_iterations`` -- the number of steps of Rauzy induction

        - ``colors`` -- (optional) colors for the towers

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: p = iet.Permutation('A B', 'B A')
            sage: T = iet.IntervalExchangeTransformation(p, [0.41510826, 0.58489174])
            sage: T.plot_towers(iterations=5)   # not tested (problem with matplotlib font cache)
            Graphics object consisting of 65 graphics primitives
        """
        px,py = map(float, position)

        T,_,towers = self.rauzy_move(iterations=iterations,data=True)
        pi = T.permutation()
        A = pi.alphabet()
        lengths = map(float, T.lengths())

        if colors is None:
            from sage.plot.colors import rainbow
            colors = {a:z for a,z in zip(A, rainbow(len(A)))}

        from sage.plot.graphics import Graphics
        from sage.plot.line import line2d
        from sage.plot.polygon import polygon2d
        from sage.plot.text import text
        G = Graphics()
        x = px
        for letter in pi[0]:
            y = x + lengths[A.rank(letter)]
            tower = towers.image(letter)
            h = tower.length()
            G += line2d([(x,py),(x,py+h)], color='black')
            G += line2d([(y,py),(y,py+h)], color='black')
            for i,a in enumerate(tower):
                G += line2d([(x,py+i),(y,py+i)], color='black')
                G += polygon2d([(x,py+i),(y,py+i),(y,py+i+1),(x,py+i+1)], color=colors[a], alpha=0.4)
                G += text(a, ((x+y)/2, py+i+.5), color='darkgray')
            G += line2d([(x,py+h),(y,py+h)], color='black', linestyle='dashed')
            G += text(letter, ((x+y)/2, py+h+.5), color='black', fontsize='large')
            x = y
        x = px
        G += line2d([(px,py-.5),(px+sum(lengths),py-.5)], color='black')
        for letter in pi[1]:
            y = x + lengths[A.rank(letter)]
            G += line2d([(x,py-.7),(x,py-.3)], color='black')
            G += text(letter, ((x+y)/2, py-.5), color='black', fontsize='large')
            x = y
        G += line2d([(x,py-.7),(x,py-.3)], color='black')

        return G

    plot = plot_two_intervals

    def show(self):
        r"""
        Shows a picture of the interval exchange transformation

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: phi = QQbar((sqrt(5)-1)/2)
            sage: t = iet.IntervalExchangeTransformation(('a b','b a'),[1,phi])
            sage: t.show() # not tested (problem with matplotlib font cache)
        """
        self.plot_two_intervals().show(axes=False)







#encoding=utf8
r"""
List of common origamis

Usage
=====

To see a list of all origami constructors, type "origamis." or "pillowcases."
and then press the tab key. The documentation for each constructor includes
information about each origami, which provides a useful reference.

Organization
============

The constructors available in this database are organized as follows

- :meth:`Escalators <OrigamiGenerators.Escalator>`
- :meth:`EierlegendeWollmilchsau <OrigamiGenerators.EierlegendeWollmilchsau>`
- :meth:`Podium <OrigamiGenerators.Podium>`
- :meth:`Stair <OrigamiGenerators.Stair>`
- :meth:`ProjectiveLine <OrigamiGenerators.ProjectiveLine>`
"""
# ****************************************************************************
#       Copyright (C) 2011-2019 Vincent Delecroix <20100.delecroix@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from __future__ import print_function, absolute_import
from six.moves import range, map, filter, zip

from .origami import Origami
from sage.rings.integer import Integer
from sage.rings.integer_ring import ZZ

class OrigamiGenerators():
    r"""
    Examples of origamis.
    """
    def __repr__(self):
        r"""
        String representation
        """
        return "Origami generators"

    @staticmethod
    def Escalator(n):
        r"""
        Escalator origamis

        The escalator origami has 2n squares and is defined by

        r = (1 2)(3 4) ... (2n-1 2n)
        u = (2 3)(4 5) ... (2n 1)

        EXAMPLES::

            sage: from surface_dynamics import *
            sage: o = origamis.Escalator(3)
            sage: o
            Escalator with 3 steps
            sage: print(str(o))
            (1,2)(3,4)(5,6)
            (1,6)(2,3)(4,5)
            sage: o.veech_group().index()
            3
        """
        lr = [(i,i+1) for i in range(1,2*n+1,2)]
        lu = [(1,2*n)] + [(i,i+1) for i in range(2,2*n,2)]
        positions = [((i+1)//2,i//2) for i in range(2*n)]
        name = "Escalator with %d steps" %n
        o = Origami(lr,lu,positions=positions,name=name)
        return o

    @staticmethod
    def EierlegendeWollmilchsau():
        r"""
        Eierlegende Wollmilchsau origami.

        This origami is associated to the quaternion group and was discovered by
        Gaby Schmithüsen.

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: o = origamis.EierlegendeWollmilchsau()
            sage: o
            Eierlegende Wollmilchsau
            sage: print(str(o))
            (1,2,3,4)(5,6,7,8)
            (1,5,3,7)(2,8,4,6)
            sage: o.veech_group().index()
            1
            sage: o.sum_of_lyapunov_exponents()
            1

            sage: oo = origamis.CyclicCover([1,1,1,1])
            sage: o.is_isomorphic(oo)
            True
        """
        r = [1,2,3,0,5,6,7,4]
        u = [4,7,6,5,2,1,0,3]
        positions = [(0,0),(1,0),(2,0),(3,0),(0,1),(1,1),(2,1),(3,1)]
        name = "Eierlegende Wollmilchsau"
        o = Origami(r,u,check=False,as_tuple=True,positions=positions,name=name)
        return o

    @staticmethod
    def CyclicCover(a, M=None):
        r"""
        Return the Abelian differential associated to the quadruple ``a``
        
        INPUT:
            
            - ``a`` - quadruple of integers
            
            - ``M`` - positive integer (default: None)

        EXAMPLES:

            sage: from surface_dynamics import *

        There are two well known cyclic covers with completely degenerated
        Lyapunov spectrum::

            sage: o = origamis.CyclicCover([1,1,1,1])
            sage: o.sum_of_lyapunov_exponents()
            1

            sage: o = origamis.CyclicCover([1,1,1,3])
            sage: o.sum_of_lyapunov_exponents()
            1

        Some examples with Veech group SL(2,Z) (this happen if and only
        if the entries of the vector is of the form ``[x,x,x,y]``)::

            sage: for a in [[1,1,1,5],[1,1,1,7],[1,3,3,3],[1,1,1,9],[1,1,1,11],[3,3,3,5]]:
            ....:     o = origamis.CyclicCover(a)
            ....:     assert o.veech_group().index() == 1
            ....:     print(o.nb_squares(), o.stratum_component())
            16 H_7(3^4)^c
            20 H_9(4^4)^odd
            20 H_9(4^4)^odd
            24 H_10(5^3, 1^3)^c
            28 H_13(6^4)^odd
            28 H_13(6^4)^odd
        """
        from sage.arith.all import gcd

        if M is None:
            M = sum(a)

        a = list(map(Integer,a))
        if len(a) != 4:
            raise ValueError("a should be of length 4")
            
        if any(ai%2==0 for ai in a):
            raise ValueError("ai should be odd")
            
        if gcd([M]+a) != 1:
            raise ValueError("gcd(M,a) should be 1")
            
        if M%2:
            raise ValueError("M should be even")

        if sum(a) % M:
            raise ValueError("the sum of ai should be 0 mod M")
            
        r = [None] * (2*M)
        u = [None] * (2*M)
        
        # initialize horizontal cylinders
        i = 0
        seen = set([])
        pos = set([0])
        neg = set([])
        while pos or neg:
            # initialize pos 2i and 2i+1
            if pos:
                i = pos.pop()

                seen.add(i)
                r[2*i] = 2*i+1
                r[2*i+1] = (2*(i+a[0]+a[3])) % (2*M)  # sigma_A + sigma_D
                j = (i + a[0] + a[3]) % M
                if j not in seen and j not in pos:
                    pos.add(j)
        
                u[2*i] = (2*(i-a[3])+1) % (2*M) # sigma_D^-1
                j = (i-a[3]) % M
                if j not in seen and j not in neg:
                    neg.add(j)
                u[2*i+1] = (2*(i+a[3])) % (2*M) # sigma_D
                j = (i+a[3]) % M
                if j not in seen and j not in neg:
                    neg.add(j)
            
            if neg:
                i = neg.pop()
                seen.add(i)
                r[2*i+1] = 2*i
                r[2*i] = (2*(i+a[1]+a[2]) + 1) % (2*M) # sigma_B + sigma_C
                j = (i+a[1]+a[2]) % M
                if j not in seen and j not in neg:
                    neg.add(j)
                    
                u[2*i] = (2*(i+a[2]) + 1) % (2*M) # sigma_C
                j = (i+a[2]) % M
                if j not in seen and j not in neg:
                    pos.add(j)
                u[2*i+1] = (2*(i-a[2])) % (2*M) # sigma_C^-1
                j = (i-a[2]) %M
                if j not in seen and j not in neg:
                    pos.add(j)
        
        return Origami(r,u,as_tuple=True,name="M_%d(%d,%d,%d,%d)" %(M,a[0],a[1],a[2],a[3]))

    @staticmethod
    def Stair(n):
        r"""
        Stair origamis

        The stair origami with n=2k squares is defined by the permutations 

        r = (1 2)(3 4) ... (n-1 n)
        u = (1)(2 3) ... (n-2 n-1)(n)

        The stair origamis with n=2k+1 squares is defined by the permutations

        r = (1 2)(3 4) ... (n-2 n-1)(n)
        u = (1)(2 3) ... (n-1 n-2)

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: o = origamis.Stair(4)
            sage: o
            Stair origami with 4 squares
            sage: print(str(o))
            (1,2)(3,4)
            (1)(2,3)(4)
            sage: o = origamis.Stair(5)
            sage: o
            Stair origami with 5 squares
            sage: print(str(o))
            (1,2)(3,4)(5)
            (1)(2,3)(4,5)
        """
        r = []
        u = [0]

        for i in range(1,n-1,2):
            r.append(i)
            r.append(i-1)
            u.append(i+1)
            u.append(i)

        if n%2:
            r.append(n-1)
        else:
            r.append(n-1)
            r.append(n-2)
            u.append(n-1)

        positions = [((i+1)//2,i//2) for i in range(n)]
        o = Origami(r,u,as_tuple=True,positions=positions,name="Stair origami with %d squares" %n)
        return o

    @staticmethod
    def Podium(data):
        r"""
        If ``data`` is an integer than the standard podium with ``data`` steps is
        returned. Otherwise, ``data`` should be a weakly decreasing list of integers
        (i.e. a integer partition).

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: o = origamis.Podium([3,3,2,1])
            sage: o
            Podium origami with partition [3, 3, 2, 1]
            sage: print(o)
            (1,2,3)(4,5,6)(7,8)(9)
            (1,4,7,9)(2,5,8)(3,6)
        """
        from sage.combinat.partition import Partition

        if isinstance(data, (int,Integer)):
            p = Partition([i for i in range(data,0,-1)])
        else:
            p = Partition(data)

        p = Partition(data)
        q = p.conjugate()

        r=[]
        positions = []
        i = 0
        for j,jj in enumerate(p):
            r.extend(range(i+1,i+jj))
            r.append(i)
            i += jj
            positions.extend((k,j) for k in range(jj))

        u = [None]*sum(p)
        for j in range(len(q)):
            k = j
            for jj in range(q[j]-1):
                u[k] = k+p[jj]
                k += p[jj]
            u[k] = j

        return Origami(r,u,positions=positions,name="Podium origami with partition %s" %str(p),as_tuple=True)

    @staticmethod
    def ProjectiveLine(p, r=None, u=None):
        r"""
        Return the projective line with action given by the matrices ``r``
        and ``u`` on the projective line over the field with ``p`` elements.

        If ``r`` and ``u`` are None, then r and u are choosen to be

        .. MATH::

            r: z \mapsto z+1   \qquad    u: z \mapsto 1/z

        The monodromy of this origami is included in `PGL(2,\ZZ/p\ZZ)`.

        EXAMPLES::

            sage: from surface_dynamics import *

        By default, the matrix used to generate the projective line origami are
        given by `z -> z+1` and `z -> 1/z` which gives a family of origamis
        which belongs to the strata ``H(2^k)``::

            sage: o = origamis.ProjectiveLine(3); o
            Projective line origami on GF(3)
             r = [1 1]    u = [0 1]
                 [0 1]        [1 0]
            sage: o.stratum()
            H_2(2)
            sage: o.veech_group().index()
            9
            sage: o.sum_of_lyapunov_exponents()
            4/3

            sage: o = origamis.ProjectiveLine(5); o
            Projective line origami on GF(5)
             r = [1 1]    u = [0 1]
                 [0 1]        [1 0]
            sage: o.stratum()
            H_3(2^2)
            sage: o.veech_group().index()
            9
            sage: o.sum_of_lyapunov_exponents()
            5/3

            sage: o = origamis.ProjectiveLine(7); o
            Projective line origami on GF(7)
             r = [1 1]    u = [0 1]
                 [0 1]        [1 0]
            sage: o.stratum()
            H_3(2^2)
            sage: o.veech_group().index()
            45
            sage: o.sum_of_lyapunov_exponents()
            5/3

        Any invertible matrix mod p can be choosed to generate the origami::

            sage: r = matrix([[1,3],[0,1]])
            sage: u = matrix([[1,0],[3,1]])
            sage: o = origamis.ProjectiveLine(5,r,u); o
            Projective line origami on GF(5)
             r = [1 3]    u = [1 0]
                 [0 1]        [3 1]
            sage: o.veech_group().index()
            10
            sage: o.sum_of_lyapunov_exponents()
            9/5
            sage: o.stratum_component()
            H_3(4)^hyp
        """
        from sage.arith.all import is_prime
        from sage.rings.finite_rings.finite_field_constructor import GF
        from sage.groups.matrix_gps.linear import GL
        from sage.modules.free_module import VectorSpace

        p = ZZ(p)
        if not p.is_prime():
            raise ValueError("p (={}) must be a prime number".format(p))

        G = GL(2,GF(p))
        V = VectorSpace(GF(p),2)

        if r is None:
            mr = G([[1,1],[0,1]])
        else:
            mr = G(r)
        if u is None:
            mu = G([[0,1],[1,0]])
        else:
            mu = G(u)

        sr = str(mr).split('\n')
        su = str(mu).split('\n')

        mr = mr
        mu = mu

        if r is None and u is None:
            positions = [(i,0) for i in range(p+1)]
        else:
            positions = None

        r = []
        u = []
        for i in range(p):
            v = V((i,1)) * mr
            if v[1]:
                r.append(int(v[0]/v[1]))
            else:
                r.append(p)

            v = V((i,1)) * mu
            if v[1]:
                u.append(int(v[0]/v[1]))
            else:
                u.append(p)

        v = V((1,0)) * mr
        if v[1]:
            r.append(int(v[0]/v[1]))
        else:
            r.append(p)

        v = V((1,0)) * mu
        if v[1]:
            u.append(int(v[0]/v[1]))
        else:
            u.append(p)

        o = Origami(r,u,
            as_tuple=True,
            positions=positions,
            name="Projective line origami on GF(%d)\n r = %s    u = %s\n     %s        %s"%(p,sr[0],su[0],sr[1],su[1]))
        return o

    @staticmethod
    def Heisenberg(p):
        r"""
        Return the Heisenberg origami.

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: h2 = origamis.Heisenberg(2)
            sage: h2.stratum_component()
            H_3(1^4)^c
            sage: h2.veech_group().index()
            3
            sage: h2.sum_of_lyapunov_exponents()
            2

            sage: h3 = origamis.Heisenberg(3)
            sage: h3.stratum_component()
            H_10(2^9)^even
            sage: h3.veech_group().index()
            1
            sage: h3.sum_of_lyapunov_exponents()
            5

            sage: h5 = origamis.Heisenberg(5)
            sage: h5.stratum_component()
            H_51(4^25)^even
            sage: h5.veech_group().index()
            1
            sage: h5.sum_of_lyapunov_exponents()
            15
        """
        p = ZZ(p)
        if not p.is_prime():
            raise ValueError("p (={}) must be prime".format(p))
        N = p**3  # map (a,b,d) -> a*p**2 + b*p + d
        pp = p*p
        r = [None]*N
        u = [None]*N
        for a in range(p):
            for b in range(p):
                for d in range(p):
                    n = a*pp + b*p + d
                    r[n] = ((a+1)%p)*pp + b*p + d
                    u[n] = a*pp + ((b+1)%p)*p + ((a+d)%p)
        return Origami(r,u,
                as_tuple=True,
                name="Heisenberg origami on GF(%d)"%p)

    def ShresthaWang(self):
        r"""
        An origami with Veech group SL(2,Z) that is not normal.

        This is the second example page 9 of https://arxiv.org/abs/1902.08214

        EXAMPLES::

            sage: from surface_dynamics import origamis
            sage: o = origamis.ShresthaWang()
            sage: o.nb_squares()
            16
            sage: o
            (1,2,3,4,5,6)(7,8,9,10,11,12)(13,14)(15,16)
            (1,8,5,15,13,12)(2,16,14,9,6,7)(3,10)(4,11)
            sage: o.stratum_component()
            H_4(2^3)^odd
            sage: o.veech_group().index()
            1
            sage: o.automorphism_group().cardinality()
            1

        And this example is not a cyclic cover::

            sage: o.is_isomorphic(origamis.CyclicCover([1,1,1,5]))
            False
        """
        return Origami("(1,2,3,4,5,6)(7,8,9,10,11,12)(13,14)(15,16)",
                       "(1,8,5,15,13,12)(2,16,14,9,6,7)(3,10)(4,11)")

origamis = OrigamiGenerators()



r"""
Fat graph.

This module is experimental.
"""
#*****************************************************************************
#       Copyright (C) 2019 Vincent Delecroix <20100.delecroix@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
#*****************************************************************************

from __future__ import absolute_import, print_function
from six.moves import range, map, zip

from sage.misc.cachefunc import cached_function
from sage.rings.integer_ring import ZZ
from sage.rings.rational_field import QQ

from array import array
from collections import deque
from surface_dynamics.misc.permutation import *

###########################
# Miscellaneous functions #
###########################

def num_and_weighted_num(it):
    from sage.rings.integer_ring import ZZ
    from sage.rings.rational_field import QQ
    s = QQ.zero()
    n = ZZ.zero()
    for _,aut in it:
        n += ZZ.one()
        if aut is None:
            s += QQ.one()
        else:
            s += QQ((1,aut.group_cardinality()))
    return n,s

def list_extrems(l, n):
    if not n:
        raise ValueError
    vdmin = vdmax = l[0]
    for i in range(1, n):
        if l[i] > vdmax:
            vdmax = l[i]
        if l[i] < vdmin:
            vdmin = l[i]
    return (vdmin, vdmax)

#####################
# Fat graph #
#####################
#
# For Abelian strata we should use constellations (= bipartite stuff)

class FatGraph(object):
    r"""
    EXAMPLES:

    The once punctured torus::

        sage: from surface_dynamics import FatGraph

        sage: vp = '(0,2,1,3)'
        sage: ep = '(0,1)(2,3)'
        sage: fp = '(0,2,1,3)'
        sage: FatGraph(vp, ep, fp)
        FatGraph('(0,2,1,3)', '(0,1)(2,3)', '(0,2,1,3)')

    Actually it is enough to specify 2 of the 3 permutations::

        sage: vp = '(0,3,1)(4)(2,5,6,7)'
        sage: ep = '(0,1)(2,3)(4,5)(6,7)'
        sage: fp = '(0,3,7,5,4,2)(1)(6)'
        sage: F0 = FatGraph(vp=vp, ep=ep, fp=fp)
        sage: F1 = FatGraph(ep=ep, fp=fp)
        sage: F2 = FatGraph(vp=vp, fp=fp)
        sage: F3 = FatGraph(vp=vp, ep=ep)
        sage: F0 == F1 and F0 == F2 and F0 == F3
        True
    """
    __slots__ = ['_n',  # number of darts (non-negative integer)
                 '_vp', # vertex permutation (array of length _n)
                 '_ep', # edge permutation (array of length _n)
                 '_fp', # face permutation (array of length _n)
                 # labels
                 # TODO: add _el and care about folded edges!!
                 '_vl', # vertex labels (array of length _n)
                 '_fl', # face labels (array of length _n)
                 # numbers
                 # TODO: add _ne
                 '_nv', # number of vertices (non-negative integer)
                 '_nf', # number of faces (non-negative integer)
                 # degrees
                 # TODO: add _ed
                 '_vd', # vertex degrees (array of length _nv)
                 '_fd'] # face degrees (array of length _nf)

    def __init__(self, vp=None, ep=None, fp=None, max_num_dart=None, check=True):
        vp, ep, fp = constellation_init(vp, ep, fp)
        self._vp = vp
        self._ep = ep
        self._fp = fp
        if len(vp) != len(ep) or len(vp) != len(fp):
            raise ValueError("invalid permutations")
        self._n = len(vp)   # number of darts
        self._nf = 0        # number of faces

        self._vl, self._vd = perm_dense_cycles(vp, self._n)
        self._nv = len(self._vd) # number of vertices
        self._fl, self._fd = perm_dense_cycles(fp, self._n)
        self._nf = len(self._fd) # number of faces

        if max_num_dart is not None:
            if max_num_dart < self._n:
                raise ValueError
            self._realloc(max_num_dart)

        if check:
            self._check()

    def __hash__(self):
        raise TypeError("FatGraph not hashable")

    def _realloc(self, max_num_dart):
            if max_num_dart < self._n:
                return
            self._vp.extend([-1] * (max_num_dart - self._n))
            self._ep.extend([-1] * (max_num_dart - self._n))
            self._fp.extend([-1] * (max_num_dart - self._n))
            self._vl.extend([-1] * (max_num_dart - self._n))
            self._fl.extend([-1] * (max_num_dart - self._n))
            self._vd.extend([-1] * (max_num_dart - self._nv))
            self._fd.extend([-1] * (max_num_dart - self._nf))

    def copy(self):
        """
        EXAMPLES::

            sage: from surface_dynamics import FatGraph
            sage: F = FatGraph.from_unicellular_word([0,1,0,2,3,4,1,4,3,2])
            sage: G = F.copy()
            sage: G._check()
        """
        F = FatGraph.__new__(FatGraph)
        F._vp = self._vp[:]
        F._ep = self._ep[:]
        F._fp = self._fp[:]
        F._n = self._n
        F._nf = self._nf
        F._nv = self._nv
        F._vl = self._vl[:]
        F._vd = self._vd[:]
        F._fl = self._fl[:]
        F._fd = self._fd[:]
        return F

    @staticmethod
    def from_unicellular_word(X):
        r"""
        Build a fat graph from a word on the letters {0, ..., n-1} where
        each letter appears exactly twice.

        EXAMPLES::

            sage: from surface_dynamics import FatGraph
            sage: FatGraph.from_unicellular_word([0,1,0,2,3,4,1,4,3,2])
            FatGraph('(0,3)(1,2,6,7)(4,9)(5,8)', '(0,2)(1,6)(3,9)(4,8)(5,7)', '(0,1,2,3,4,5,6,7,8,9)')
            sage: FatGraph.from_unicellular_word([0,1,2,0,3,2,4,1,3,4])
            FatGraph('(0,6,2,7,9,4)(1,3,5,8)', '(0,3)(1,7)(2,5)(4,8)(6,9)', '(0,1,2,3,4,5,6,7,8,9)')
        """
        n = len(X)
        m = n // 2
        ep = [None] * n
        vp = [None] * n
        fp = list(range(1,n)) + [0]
        symb_to_pos = [None] * m
        for i,k in enumerate(X):
            j = symb_to_pos[k]
            if j is not None:
                ep[i] = j
                ep[j] = i
                vp[(j + 1) % n] = i
                vp[(i + 1) % n] = j
            else:
                symb_to_pos[k] = i
        return FatGraph(vp, ep, fp)

    @staticmethod
    def from_string(s):
        r"""
        Build a fat graph from a serialized string.

        See also :meth:`to_string`.

        EXAMPLES::

            sage: from surface_dynamics import FatGraph
            sage: s = '20_i23017546b98jchedfag_2301547698badcfehgji_12346758ab9igdhfejc0'
            sage: F = FatGraph.from_string(s)
            sage: F.to_string() == s
            True
            sage: FatGraph.from_string('0___')
            FatGraph('()', '()', '()')
        """
        if not isinstance(s, str) or s.count('_') != 3:
            raise ValueError("invalid input")
        n, vp, ep, fp = s.split('_')
        n = int(n)
        vp = perm_from_base64_str(vp, n)
        ep = perm_from_base64_str(ep, n)
        fp = perm_from_base64_str(fp, n)
        return FatGraph(vp, ep, fp)

    def to_string(self):
        r"""
        Serialization to string.

        EXAMPLES::

            sage: from surface_dynamics import FatGraph
            sage: FatGraph.from_unicellular_word([0,1,0,2,3,4,1,4,3,2]).to_string()
            '10_3260987154_2609871543_1234567890'
            sage: FatGraph('', '', '').to_string()
            '0___'
        """
        n = self._n
        return str(n) + "_" + \
               perm_base64_str(self._vp, n) + "_" + \
               perm_base64_str(self._ep, n) + "_" + \
               perm_base64_str(self._fp, n)

    def _check(self, error=RuntimeError):
        vp = self._vp
        vl = self._vl
        vd = self._vd

        ep = self._ep

        fp = self._fp
        fl = self._fl
        fd = self._fd

        n = self._n
        nf = self._nf
        nv = self._nv

        m = sum(vp[i] != -1 for i in range(n))

        if not perm_check(vp, n):
            raise ValueError("invalid vertex permutation: %s" % vp)
        if not perm_check(ep, n):
            raise ValueError("invalid edge permutation: %s" % ep)
        if not perm_check(fp, n):
            raise ValueError("invalid face permutation: %s" % fp)

        if perm_num_cycles(vp, n) != self._nv:
            raise error("wrong number of vertices")
        if perm_num_cycles(fp, n) != self._nf:
            raise error("wrong number of faces")

        if len(vl) < n or len(fl) < n or len(vd) < nv or len(fd) < nf:
               raise error("inconsistent lengths")

        if any(x < 0 or x > n for x in vd[:nv]) or sum(vd[:nv]) != m:
            raise error("invalid vertex degrees")
        if any(x < 0 or x > n for x in fd[:nf]) or sum(fd[:nf]) != m:
            raise error("invalid face degrees")

        ffd = [0] * nf
        vvd = [0] * nv

        for i in range(n):
            if vp[i] == -1:
                if ep[i] != -1 or fp[i] != -1:
                    raise ValueError("inconsistent dart activity for i={}".format(i))
                continue
            elif ep[i] == -1 or fp[i] == -1:
                raise ValueError("inconsistent dart activity for i={}".format(i))

            if fp[ep[vp[i]]] != i:
                raise error("fp[ep[vp[%d]]] = %d" % (i, fp[ep[vp[i]]]))
            if fl[i] < 0 or fl[i] >= nf:
                raise error("face label out of range: fl[%d] = %d" % (i, fl[i]))
            if vl[i] < 0 or vl[i] >= nv:
                raise error("vertex label out of range: vl[%d] = %d" % (i, vl[i]))

            if fl[fp[i]] != fl[i]:
                raise error("fl[fp[%d]] = %d while fl[%d] = %d" %(i, fl[fp[i]], i, fl[i]))
            
            if vl[vp[i]] != vl[i]:
                raise error("vl[vp[%d]] = vl[%d] = %d while vl[%d] = %d" %(i, vp[i], vl[vp[i]], i, vl[i]))

            ffd[fl[i]] += 1
            vvd[vl[i]] += 1

        if vvd != vd[:nv]:
            raise error("inconsistent face labels/degrees, got %s instead of vd = %s" % (vvd, vd[:nv]))
        if ffd != fd[:nf]:
            raise error("inconsistent vertex labels/degrees, got %s instead of fd = %s" %(ffd, fd[:nf]))

    def is_face_bipartite(self):
        r"""
        Test whether the faces admit a bi-coloring.

        EXAMPLES::

            sage: from surface_dynamics.topology.fat_graph import *
            sage: vp = '(0,2,1,3)'
            sage: ep = '(0,1)(2,3)'
            sage: fp = '(0,2,1,3)'
            sage: F = FatGraph(vp, ep, fp, 6)
            sage: F.is_face_bipartite()
            False
            sage: F.split_face(0,1)
            sage: F.is_face_bipartite()
            True

            sage: vp = '(0,5,1,2,3,4)'
            sage: ep = '(0,2)(1,3)(4,5)'
            sage: fp = '(0,1,2,4)(3,5)'
            sage: FatGraph(vp, ep, fp).is_face_bipartite()
            False

            sage: from surface_dynamics.topology.fat_graph_exhaustive_generation import FatGraphs_g_nf_nv
            sage: F = FatGraphs_g_nf_nv(1, 3, 3, vertex_min_degree=3)
            sage: F.cardinality_and_weighted_cardinality(filter=lambda x,a: x.is_face_bipartite())
            (3, 5/3)
        """
        # trivial cases
        if self._nf == 0:
            return True
        elif self._nf == 1:
            return False

        n = self._n
        ep = self._ep
        fp = self._fp
        fl = self._fl
        nf = self._nf
        colors = [-1] * nf
        edge_seen = [False] * n
        to_test = perm_orbit(fp, 0)
        colors[self._fl[0]] = 1
        while to_test:
            e1 = to_test.pop()
            if edge_seen[e1]:
                continue
            e2 = ep[e1]
            f1 = fl[e1]
            f2 = fl[e2]
            if colors[f1] == -1:
                raise RuntimeError
            elif colors[f2] == -1:
                # discover a new face
                colors[f2] = 1 - colors[f1]
                to_test.extend(perm_orbit(fp, e2))
            elif colors[f1] == colors[f2]:
                # contradiction in colors
                return False

            edge_seen[e1] = edge_seen[e2] = True

        return True

    def __copy__(self):
        r"""
        EXAMPLES::

            sage: from surface_dynamics.topology.fat_graph import FatGraph

            sage: vp = '(0,2,1,3)'
            sage: ep = '(0,1)(2,3)'
            sage: fp = '(0,2,1,3)'
            sage: cm = FatGraph(vp, ep, fp)
            sage: cm2 = cm.__copy__()
            sage: cm2._check()
        """
        cm = FatGraph.__new__(FatGraph)
        cm._vp = self._vp[:]
        cm._ep = self._ep[:]
        cm._fp = self._fp[:]
        cm._n = self._n
        cm._nf = self._nf
        cm._nv = self._nv
        cm._vl = self._vl[:]
        cm._fl = self._fl[:]
        cm._vd = self._vd[:]
        cm._fd = self._fd[:]
        return cm

    def __repr__(self):
        n = self._n
        fd = self._fd[:self._nf]
        vd = self._vd[:self._nv]
        fd.sort(reverse=True)
        vd.sort(reverse=True)
        return "FatGraph('%s', '%s', '%s')" % (perm_cycle_string(self._vp, True, n),
                                               perm_cycle_string(self._ep, True, n),
                                               perm_cycle_string(self._fp, True, n))

    def __eq__(self, other):
        r"""
        TESTS::

            sage: from surface_dynamics.topology.fat_graph import FatGraph

            sage: vp = '(0,2,1,3)'
            sage: ep = '(0,1)(2,3)'
            sage: fp = '(0,2,1,3)'
            sage: cm1 = FatGraph(vp, ep, fp)
            sage: cm2 = FatGraph(vp, ep, fp, 100)
            sage: cm1 == cm2
            True
        """
        if type(self) != type(other):
            raise TypeError

        if self._n != other._n or self._nf != other._nf or self._nv != other._nv:
            return False

        for i in range(self._n):
            if self._vp[i] != other._vp[i] or \
               self._ep[i] != other._ep[i] or \
               self._fp[i] != other._fp[i]:
                   return False

        # here we ignore the vertex and face labels...
        return True

    def __ne__(self, other):
        return not self == other

    def vertex_permutation(self, copy=True):
        if copy:
            return self._vp[:self._n]
        else:
            return self._vp

    def is_trivalent(self):
        vp = self._vp
        for i in range(self._n):
            if vp[i] == i or vp[vp[i]] == i or vp[vp[vp[i]]] != i:
                return False
        return True

    def edge_permutation(self, copy=True):
        if copy:
            return self._ep[:self._n]
        else:
            return self._ep

    def face_permutation(self, copy=True):
        if copy:
            return self._fp[:self._n]
        else:
            return self._fp

    def vertex_profile(self):
        return perm_cycle_type(self._vp, self._n)

    def edge_profile(self):
        return perm_cycle_type(self._ep, self._n)

    def face_profile(self):
        return perm_cycle_type(self._fp, self._n)

    def profile(self):
        return (self.vertex_profile(), self.edge_profile(), self.face_profile())

    def num_darts(self):
        return self._n

    def num_folded_edges(self):
        return sum(self._ep[i] == 1 for i in range(self._n))

    def num_faces(self):
        return self._nf

    def num_vertices(self):
        return self._nv

    def vertices(self):
        r"""
        EXAMPLES::

            sage: from surface_dynamics.topology.fat_graph import FatGraph
            sage: vp = '(0,4,2,5,1,3)'
            sage: ep = '(0,1)(2,3)(4,5)'
            sage: fp = '(0,5)(1,3,4,2)'
            sage: cm = FatGraph(vp, ep, fp)
            sage: cm.vertices()
            [[0, 4, 2, 5, 1, 3]]
        """
        return perm_cycles(self._vp, True, self._n)

    def vertex_degrees(self):
        return self._vd[:self._nv]

    def vertex_degree_extrems(self):
        return list_extrems(self._vd, self._nv)

    def vertex_degree_min(self):
        return list_extrems(self._vd, self._nv)[0]
    
    def vertex_degree_max(self):
        return list_extrems(self._vd, self._nv)[1]

    def face_degree_extrems(self):
        return list_extrems(self._fd, self._nf)

    def face_degree_min(self):
        return list_extrems(self._fd, self._nf)[0]

    def face_degree_max(self):
        return list_extrems(Self._fd, self._nf)[1]

    def face_degree_min(self):
        return s

    def edges(self):
        r"""
        EXAMPLES::

            sage: from surface_dynamics.topology.fat_graph import FatGraph
            sage: vp = '(0,4,2,5,1,3)'
            sage: ep = '(0,1)(2,3)(4,5)'
            sage: fp = '(0,5)(1,3,4,2)'
            sage: cm = FatGraph(vp, ep, fp)
            sage: cm.edges()
            [[0, 1], [2, 3], [4, 5]]
        """
        return perm_cycles(self._ep, True, self._n)

    def num_edges(self):
        return self._n // 2

    def faces(self):
        r"""
        EXAMPLES::

            sage: from surface_dynamics.topology.fat_graph import FatGraph
            sage: vp = '(0,4,2,5,1,3)'
            sage: ep = '(0,1)(2,3)(4,5)'
            sage: fp = '(0,5)(1,3,4,2)'
            sage: cm = FatGraph(vp, ep, fp)
            sage: cm.faces()
            [[0, 5], [1, 3, 4, 2]]
        """
        return perm_cycles(self._fp, True, self._n)

    def face_degrees(self):
        return self._fd[:self._nf]

    def genus(self):
        r"""
        EXAMPLES::

            sage: from surface_dynamics.topology.fat_graph import FatGraph
            sage: vp = '(0,4,2,5,1,3)'
            sage: ep = '(0,1)(2,3)(4,5)'
            sage: fp = '(0,5)(1,3,4,2)'
            sage: cm = FatGraph(vp, ep, fp)
            sage: cm.genus()
            1
        """
        return (2 - self._nf + self._n//2 - self._nv) // 2

    def euler_characteristic(self):
        r"""
        EXAMPLES::

            sage: from surface_dynamics.topology.fat_graph import FatGraph
            sage: vp = '(0,4,2,5,1,3)'
            sage: ep = '(0,1)(2,3)(4,5)'
            sage: fp = '(0,5)(1,3,4,2)'
            sage: cm = FatGraph(vp, ep, fp)
            sage: cm.euler_characteristic()
            0
        """
        return self._nf - self._n//2 + self._nv

    def dual(self):
        r"""
        Return the dual fat graph.

        EXAMPLES::

            sage: from surface_dynamics.topology.fat_graph import FatGraph
            sage: F = FatGraph(vp=None,ep='(0,1)',fp='(0)(1)')
            sage: F.dual()
            sage: F
            FatGraph('(0)(1)', '(0,1)', '(0,1)')
            sage: F._check()
            sage: s = '20_i23017546b98jchedfag_2301547698badcfehgji_12346758ab9igdhfejc0'
            sage: F = FatGraph.from_string(s)
            sage: F.dual()
            sage: F._check()
            sage: F.dual()
            sage: F._check()
            sage: F == FatGraph.from_string(s)
            True
        """
        # TODO: invert in place !!!!
        self._vp, self._fp = perm_invert(self._fp, self._n), perm_invert(self._vp, self._n)
        self._nv, self._nf = self._nf, self._nv
        self._vl, self._fl = self._fl, self._vl
        self._vd, self._fd = self._fd, self._vd

    def polytope(self, b, min_length=0):
        r"""
        Return the polytope of edge lengths where the input ``b`` specifies the
        length of faces.

        EXAMPLES::

            sage: from surface_dynamics.topology.fat_graph import FatGraph
            sage: vp = '(0,4,2,5,1,3)'
            sage: ep = '(0,1)(2,3)(4,5)'
            sage: fp = '(0,5)(1,3,4,2)'
            sage: cm = FatGraph(vp, ep, fp)

            sage: cm.polytope([3, 5], min_length=1).vertices_list()
            [[1, 1, 1, 1, 2, 2], [2, 2, 1, 1, 1, 1]]
            sage: cm.polytope([3, 5], min_length=0).vertices_list()
            [[0, 0, 1, 1, 3, 3], [3, 3, 1, 1, 0, 0]]

            sage: vp = '(0,3,4)(1,2,6)(5)(7)'
            sage: ep = '(0,1)(2,3)(4,5)(6,7)'
            sage: fp = '(0,6,7,2)(1,4,5,3)'
            sage: cm = FatGraph(vp, ep, fp)
            sage: cm.polytope([5, 7])
            A 2-dimensional polyhedron in QQ^8 defined as the convex hull of 3 vertices

        TESTS::

            sage: from surface_dynamics.topology.fat_graph import FatGraph
            sage: vp = '(0,4,2,5,1,3)'
            sage: ep = '(0,1)(2,3)(4,5)'
            sage: fp = '(0,5)(1,3,4,2)'
            sage: cm = FatGraph(vp, ep, fp)

            sage: cm.polytope([3, 5, 2])
            Traceback (most recent call last):
            ...
            ValueError: the length of b must be the number of faces
        """
        if self.num_faces() != len(b):
            raise ValueError("the length of b must be the number of faces")

        ep = self._ep
        fp = self._fp

        ieqs = []
        # positivity
        for i in range(self._n):
            l = [0] * self._n
            l[i] = 1
            ieqs.append([-min_length] + l)

        eqns = []
        # half edge equations
        for i,j in self.edges():
            l = [0] * self._n
            l[i] = 1
            l[j] = -1
            eqns.append([0] + l)

        # face equations
        for bb,f in zip(b,self.faces()):
            l = [0] * self._n
            for i in f:
                l[i] = 1
            eqns.append([-bb] + l)

        from sage.geometry.polyhedron.constructor import Polyhedron
        return Polyhedron(ieqs=ieqs, eqns=eqns)

    def integral_points(self, b, min_length=1):
        r"""
        Return the edge lengths solution to the face lengths constraint ``b``.

        EXAMPLES::

            sage: from surface_dynamics.topology.fat_graph import FatGraph
            sage: vp = '(0,4,2,5,1,3)'
            sage: ep = '(0,1)(2,3)(4,5)'
            sage: fp = '(0,5)(1,3,4,2)'
            sage: cm = FatGraph(vp, ep, fp)
            sage: cm.integral_points((2,4))
            ((1, 1, 1, 1, 1, 1),)
            sage: cm.integral_points((3,5))
            ((1, 1, 1, 1, 2, 2), (2, 2, 1, 1, 1, 1))
            sage: cm.integral_points((5,11))
            ((1, 1, 3, 3, 4, 4),
             (2, 2, 3, 3, 3, 3),
             (3, 3, 3, 3, 2, 2),
             (4, 4, 3, 3, 1, 1))
        """
        return self.polytope(b, min_length).integral_points()

#    def kontsevich_volume_rational_function(self, R=None):
#        r"""
#        This is not under an appropriate form...
#        """
#        from sage.rings.rational_field import QQ
#        from sage.rings.polynomial.polynomial_ring_constructor import PolynomialRing
#
#        nf = self._nf
#        fl = self._fl
#        n = self._n
#        ep = self._ep
#        if R is None:
#            R = PolynomialRing(QQ, 'b', nf)
#        gens = R.gens()
#        res = R.one()
#        for i in range(self._n):
#            j = ep[i]
#            if j < i:
#                continue
#            res *= 1 / self.automorphism_group().group_cardinality() / (gens[fl[i]] + gens[fl[j]])
#        return res

    ##############################
    # Augmentation and reduction #
    ##############################

    def _check_alloc(self, n, nv, nf):
        if len(self._vp) < n or \
           len(self._ep) < n or \
           len(self._fp) < n or \
           len(self._vl) < n or \
           len(self._fl) < n or \
           len(self._fd) < nf or \
           len(self._vd) < nv:
               raise TypeError("reallocation needed")

    def split_face(self, i, j):
        r"""
        Insert an edge between the darts ``i`` and ``j`` to split the face.

        One of the face will contains i, fp[i], ...,  (the x-face) and the 
        other one will contain j, fp[j], ... In the special case i=j, a
        monogon (= face with only one edge) is created.

        The converse operation is implemented in :meth:`remove_edge`.

        EXAMPLES:

        The once punctured torus::

            sage: from surface_dynamics.topology.fat_graph import FatGraph

            sage: vp = '(0,2,1,3)'
            sage: ep = '(0,1)(2,3)'
            sage: fp = '(0,2,1,3)'
            sage: eps = '(0,1)(2,3)(4,5)'

            sage: vp20 = '(0,4,2,5,1,3)'
            sage: fp20 = '(0,5)(1,3,4,2)'
            sage: cm = FatGraph(vp, ep, fp, 6)
            sage: cm.split_face(2,0)
            sage: cm == FatGraph(vp20, eps, fp20)
            True

            sage: vp10 = '(0,4,2,1,5,3)'
            sage: fp10 = '(0,2,5)(1,3,4)'
            sage: cm = FatGraph(vp, ep, fp, 6)
            sage: cm.split_face(1,0)
            sage: cm == FatGraph(vp10, eps, fp10)
            True

            sage: vp30 = '(0,4,2,1,3,5)'
            sage: fp30 = '(0,2,1,5)(3,4)'
            sage: cm = FatGraph(vp, ep, fp, 6)
            sage: cm.split_face(3,0)
            sage: cm == FatGraph(vp30, eps, fp30)
            True

            sage: vp00 = '(0,5,4,2,1,3)'
            sage: fp00 = '(0,2,1,3,4)(5)'
            sage: cm = FatGraph(vp, ep, fp, 6)
            sage: cm.split_face(0,0)
            sage: cm == FatGraph(vp00, eps, fp00)
            True

            sage: vp22 = '(0,2,5,4,1,3)'
            sage: fp22 = '(0,4,2,1,3)(5)'
            sage: cm = FatGraph(vp, ep, fp, 6)
            sage: cm.split_face(2,2)
            sage: cm == FatGraph(vp22, eps, fp22)
            True

        A genus 2 surface::

            sage: vp = '(0,3,6,8)(1,10,9,12,5)(2,7,4,11)(13)'
            sage: ep = '(0,1)(2,3)(4,5)(6,7)(8,9)(10,11)(12,13)'
            sage: fp = '(0,5,7,3,11,1,8,10,4,12,13,9,6,2)'
            sage: cm = FatGraph(vp, ep, fp, 21)
            sage: cm.split_face(0,1); cm._check()
            sage: cm.split_face(4,13); cm._check()
            sage: cm.split_face(5,14); cm._check()
            sage: cm.remove_edge(18); cm._check()
            sage: cm.remove_edge(16); cm._check()
            sage: cm.remove_edge(14); cm._check()
            sage: cm == FatGraph(vp, ep, fp, 21)
            True
        """
        vp = self._vp
        vl = self._vl
        vd = self._vd

        ep = self._ep

        fp = self._fp
        fl = self._fl
        fd = self._fd

        n = self._n
        nf = self._nf
        nv = self._nv

        i = int(i)
        j = int(j)
        if i < 0 or i >= self._n or j < 0 or j >= n or fl[i] != fl[j]:
            raise ValueError("invalid darts i=%d and j=%d for face splitting" %(i, j))

        self._check_alloc(n + 2, nv, nf + 1)

        x = self._n
        y = self._n + 1
        ii = ep[vp[i]] # = fp^-1(i)
        jj = ep[vp[j]] # = fp^-1(j)

        ep[x] = y
        ep[y] = x

        self._n += 2
        self._nf += 1

        if i == j:
            # add a monogon
            # fp (i A) -> (i A x)(y)
            # vp (... i ...) -> (... i y x ...)
            fp[ii] = x
            fp[x] = i
            fp[y] = y
            vp[x] = vp[i]
            vp[y] = x
            vp[i] = y


            vl[x] = vl[y] = vl[i]
            fl[x] = fl[i]
            fl[y] = nf

            fd[fl[i]] += 1
            fd[nf] = 1

            vd[vl[i]] += 2

        else:
            # general case
            # update permutations:
            # fp  (i A j B)               -> (i A x) (j B y)
            # ep                          -> (x y)
            # vp  (... i ...) (... j ...) -> (... i y ...) (... j x ...)
            fp[jj] = x
            fp[x] = i
            fp[ii] = y
            fp[y] = j

            vp[y] = vp[i]
            vp[i] = y
            vp[x] = vp[j]
            vp[j] = x

            # update labels and degrees
            vl[x] = vl[j]
            vl[y] = vl[i]

            fl[x] = fl[i]
            fl[y] = fl[j]

            dfy = 0  # degree of the y-face
            while fl[y] != nf:
                fl[y] = nf
                y = fp[y]
                dfy += 1
            dfx = fd[fl[x]] + 2 - dfy

            fd[fl[x]] = dfx
            fd[fl[y]] = dfy
            vd[vl[x]] += 1
            vd[vl[y]] += 1

    def remove_edge(self, i):
        r"""
        Remove an edge.

        If the edge has the same face on both sides, then the genus drops by 1.
        Inverse operation of :meth:`split_face` or :meth:`trisect_face`.

        EXAMPLES::

            sage: from surface_dynamics.topology.fat_graph import FatGraph

            sage: vp = '(0,2,1,3)'
            sage: ep = '(0,1)(2,3)'
            sage: fp = '(0,2,1,3)'
            sage: cm = FatGraph(vp, ep, fp)

            sage: eps = '(0,1)(2,3)(4,5)'
            sage: vp20 = '(0,5,4,2,1,3)'
            sage: fp20 = '(0,2,1,3,4)(5)'
            sage: cm2 = FatGraph(vp20, eps, fp20, 6)
            sage: cm2.remove_edge(4)
            sage: cm2 == cm
            True
            sage: cm2 = FatGraph(vp20, eps, fp20, 6)
            sage: cm2.remove_edge(5)
            sage: cm2 == cm
            True

            sage: vp10 = '(0,4,2,5,1,3)'
            sage: fp10 = '(0,5)(1,3,4,2)'
            sage: cm2 = FatGraph(vp10, eps, fp10)
            sage: cm2.remove_edge(4)
            sage: cm2 == cm
            True
            sage: cm2 = FatGraph(vp10, eps, fp10)
            sage: cm2.remove_edge(5)
            sage: cm2 == cm
            True

            sage: vp30 = '(0,4,2,1,5,3)'
            sage: fp30 = '(0,2,5)(1,3,4)'
            sage: cm2 = FatGraph(vp30, eps, fp30)
            sage: cm2.remove_edge(4)
            sage: cm2 == cm
            True
            sage: cm2 = FatGraph(vp30, eps, fp30)
            sage: cm2.remove_edge(5)
            sage: cm2 == cm
            True

            sage: vp00 = '(0,5,4,2,1,3)'
            sage: fp00 = '(0,2,1,3,4)(5)'
            sage: cm2 = FatGraph(vp00, eps, fp00)
            sage: cm2.remove_edge(4)
            sage: cm2 == cm
            True

            sage: vp22 = '(0,2,5,4,1,3)'
            sage: fp22 = '(0,4,2,1,3)(5)'
            sage: cm2 = FatGraph(vp00, eps, fp00)
            sage: cm2.remove_edge(4)
            sage: cm2 == cm
            True
        """
        vp = self._vp
        ep = self._ep
        fp = self._fp

        vl = self._vl
        fl = self._fl

        vd = self._vd
        fd = self._fd

        n = self._n
        nf = self._nf
        nv = self._nv

        i = int(i)
        if i < 0 or i >= self._n:
            raise ValueError("dart index out of range")
        j = ep[i]

        fi = fl[i]
        fj = fl[j]
        if fi == fj:
            raise ValueError("i=%d and j=%d on the same face" %(i,j))
        fmin = min(fi, fj)
        if i < n - 2 or j < n - 2 or max(fi, fj) != nf-1:
            raise NotImplementedError

        ii = ep[vp[i]]
        jj = ep[vp[j]]
        if fd[fl[i]] == 1:
            # monogon
            assert vp[i] == j
            fp[jj] = fp[j]
            vp[fp[j]] = vp[j]
        elif fd[fl[j]] == 1:
            # monogon
            assert vp[j] == i
            fp[ii] = fp[i]
            vp[fp[i]] = vp[i]
        else:
            # none of them are monogons
            fp[ii] = fp[j]
            fp[jj] = fp[i]

            vp[fp[j]] = vp[i]
            vp[fp[i]] = vp[j]

        # update vertex and face degrees
        vd[vl[i]] -= 1
        vd[vl[j]] -= 1

        d = fd[fl[i]] + fd[fl[j]] - 2
        fd[fmin] = d

        # update face labels
        k = fp[i]
        while fl[k] != fmin:
            fl[k] = fmin
            k = fp[k]
        k = fp[j]
        while fl[k] != fmin:
            fl[k] = fmin
            k = fp[k]

        self._n -= 2
        self._nf -= 1

    def split_vertex(self, i, j):
        r"""
        Insert a new edge to split the vertex located at the darts i and j.
        
        This operation keeps the genus constant. The inverse operation is implemented
        in :meth:`contract_edge`.

        EXAMPLES::

            sage: from surface_dynamics.topology.fat_graph import FatGraph

            sage: vp = '(0,2,1,3)'
            sage: ep = '(0,1)(2,3)'
            sage: fp = '(0,2,1,3)'
            sage: eps = '(0,1)(2,3)(4,5)'

            sage: vp02 = '(0,4,1,3)(2,5)'
            sage: fp02 = '(0,4,2,1,3,5)'
            sage: cm = FatGraph(vp, ep, fp, 6)
            sage: cm.split_vertex(0,2)
            sage: cm == FatGraph(vp02, eps, fp02)
            True

            sage: vp01 = '(0,4,3)(1,5,2)'
            sage: fp01 = '(0,2,4,1,3,5)'
            sage: cm = FatGraph(vp, ep, fp, 6)
            sage: cm.split_vertex(0,1)
            sage: cm == FatGraph(vp01, eps, fp01)
            True

            sage: vp03 = '(0,4)(1,3,5,2)'
            sage: fp03 = '(0,2,1,4,3,5)'
            sage: cm = FatGraph(vp, ep, fp, 6)
            sage: cm.split_vertex(0,3)
            sage: cm == FatGraph(vp03, eps, fp03)
            True
        """
        vp = self._vp
        ep = self._ep
        fp = self._fp
        vl = self._vl
        fl = self._fl
        n = self._n
        nf = self._nf
        nv = self._nv
        vd = self._vd
        fd = self._fd

        i = int(i)
        j = int(j)
        if i < 0 or i >= self._n or j < 0 or j >= self._n or vl[i] != vl[j]:
            raise ValueError("invalid darts i=%d and j=%d for vertex splitting" %(i, j))

        self._check_alloc(n + 2, nv + 1, nf)

        x = self._n
        y = self._n + 1
        ii = vp[i]
        jj = vp[j]
        ep[x] = y
        ep[y] = x
        self._n += 2
        self._nv += 1

        if i == j:
            # introduce a vertex of degree 1
            vp[x] = ii
            vp[i] = x
            vp[y] = y

            fp[y] = i
            fp[x] = y
            fp[ep[ii]] = x

            fl[x] = fl[y] = fl[i]
            vl[x] = vl[i]
            vl[y] = nv

            vd[vl[x]] += 1
            vd[vl[y]] = 1

            fd[fl[x]] += 2

        else:
            # general case
            # update permutations
            # fp (... i ...) (... j ...) -> (... y i ...) (... x j ...)
            # ep                         -> (x y)
            # vp (A i B j)               -> (A i x) (B j y)
            vp[x] = jj
            vp[i] = x
            vp[y] = ii
            vp[j] = y

            fp[ep[jj]] = x
            fp[x] = j
            fp[ep[ii]] = y
            fp[y] = i 

            # update labels and degrees

            fl[x] = fl[j]
            fl[y] = fl[i]

            vl[x] = vl[i]
            vl[y] = vl[j]

            dvy = 0
            while vl[y] != nv:
                vl[y] = nv
                y = vp[y]
                dvy += 1
            dvx = vd[vl[x]] + 2 - dvy

            vd[vl[x]] = dvx
            vd[vl[y]] = dvy
            fd[fl[x]] += 1
            fd[fl[y]] += 1

    def contract_edge(self, i):
        r"""
        Contract an edge between two distinct zeros.

        Inverse operation of :meth:`split_vertex` except that here we allow
        vertices of degree one.

        EXAMPLES::

            sage: from surface_dynamics.topology.fat_graph import FatGraph

            sage: vp = '(0,2,1,3)'
            sage: ep = '(0,1)(2,3)'
            sage: fp = '(0,2,1,3)'
            sage: eps = '(0,1)(2,3)(4,5)'

            sage: vp02 = '(0,4,1,3)(2,5)'
            sage: fp02 = '(0,4,2,1,3,5)'
            sage: cm = FatGraph(vp02, eps, fp02)
            sage: cm.contract_edge(4)
            sage: cm == FatGraph(vp, ep, fp)
            True
            sage: cm = FatGraph(vp02, eps, fp02)
            sage: cm.contract_edge(5)
            sage: cm == FatGraph(vp, ep, fp)
            True

            sage: vp01 = '(0,4,3)(1,5,2)'
            sage: fp01 = '(0,2,4,1,3,5)'
            sage: cm = FatGraph(vp01, eps, fp01)
            sage: cm.contract_edge(4)
            sage: cm == FatGraph(vp, ep, fp)
            True
            sage: cm = FatGraph(vp01, eps, fp01)
            sage: cm.contract_edge(5)
            sage: cm == FatGraph(vp, ep, fp)
            True

            sage: vp03 = '(0,4)(1,3,5,2)'
            sage: fp03 = '(0,2,1,4,3,5)'
            sage: cm = FatGraph(vp03, eps, fp03)
            sage: cm.contract_edge(4)
            sage: cm == FatGraph(vp, ep, fp)
            True
            sage: cm = FatGraph(vp03, eps, fp03)
            sage: cm.contract_edge(5)
            sage: cm == FatGraph(vp, ep, fp)
            True

        Degree 1 vertices::

            sage: cm = FatGraph('(0,2)(1)(3)', '(0,1)(2,3)', '(0,1,2,3)')
            sage: cm.contract_edge(2)
            sage: cm
            FatGraph('(0)(1)', '(0,1)', '(0,1)')
            sage: cm2 = FatGraph('(0,2)(1)(3)', '(0,1)(2,3)', '(0,1,2,3)')
            sage: cm2.contract_edge(3)
            sage: cm == cm2
            True
        """
        vp = self._vp
        ep = self._ep
        fp = self._fp

        vl = self._vl
        fl = self._fl

        vd = self._vd
        fd = self._fd

        n = self._n
        nf = self._nf
        nv = self._nv

        i = int(i)
        if i < 0 or i >= self._n:
            raise ValueError("dart index out of range")
        j = ep[i]
        if vl[i] == vl[j]:
            raise ValueError("i=%d and j=%d on the same vertex" %(i,j))

        vi = vl[i]
        vj = vl[j]
        vmin = min(vi, vj)
        if i < n - 2 or j < n - 2 or max(vi, vj) != nv-1:
            raise NotImplementedError

        ii = ep[vp[i]]
        jj = ep[vp[j]]
        if vd[vl[i]] == 1:
            # vertex of degree one
            assert fp[j] == i
            vp[fp[i]] = vp[j]
            fp[jj] = fp[i]

        elif vd[vl[j]] == 1:
            # vertex of degree one
            assert fp[i] == j
            vp[fp[j]] = vp[i]
            fp[ii] = fp[j]

        else:
            vp[fp[i]] = vp[i]
            vp[fp[j]] = vp[j]
            fp[ii] = fp[i]
            fp[jj] = fp[j]

        # update vertex and face degree
        fd[fl[i]] -= 1
        fd[fl[j]] -= 1

        d = vd[vl[i]] + vd[vl[j]] - 2
        vd[vmin] = d

        # update vertex labels
        k = vp[i]
        while vl[k] != vmin:
            vl[k] = vmin
            k = vp[k]
        k = vp[j]
        while vl[k] != vmin:
            vl[k] = vmin
            k = vp[k]

        self._n -= 2
        self._nv -= 1

    def trisect_face(self, i, j, k):
        r"""
        Insert a bridge

        INPUT:

        - ``i``, ``j``, ``k`` - dart in the same face in counter-clockwise
          order


        EXAMPLES::

            sage: from surface_dynamics.topology.fat_graph import FatGraph

            sage: vp = '(0,2,1,3)'
            sage: ep = '(0,1)(2,3)'
            sage: fp = '(0,2,1,3)'
            sage: cm = FatGraph(vp, ep, fp, 8)

            sage: vp021 = '(0,7,2,6,5,1,4,3)'
            sage: ep021 = '(0,1)(2,3)(4,5)(6,7)'
            sage: fp021 = '(0,5,1,3,7,2,4,6)'
            sage: cm021 = FatGraph(vp021, ep021, fp021)
            sage: cm.trisect_face(0, 2, 1)
            sage: cm == cm021
            True

            sage: cm = FatGraph(vp, ep, fp, 10)
            sage: cm.trisect_face(0, 0, 3)

            sage: cm = FatGraph(vp, ep, fp, 10)
            sage: cm.trisect_face(0, 3, 3)

            sage: cm = FatGraph(vp, ep, fp, 10)
            sage: cm.trisect_face(0, 3, 0)

            sage: cm = FatGraph(vp, ep, fp, 10)
            sage: cm.trisect_face(0, 0, 0)
        """
        vp = self._vp
        ep = self._ep
        fp = self._fp

        vl = self._vl
        fl = self._fl

        vd = self._vd
        fd = self._fd

        n = self._n
        nf = self._nf
        nv = self._nv

        i = int(i)
        j = int(j)
        k = int(k)

        if i < 0 or i >= n or j < 0 or j >= n or k < 0 or k >= n:
            raise ValueError("dart index out of range")
        if fl[i] != fl[j] or fl[i] != fl[k]:
            raise ValueError("darts in distinct faces")

        self._check_alloc(n + 4, nv, nf)
        self._n += 4

        ii = ep[vp[i]] # = fp^-1(i) at the end of B
        jj = ep[vp[j]] # = fp^-1(j) at the end of A
        kk = ep[vp[k]] # = fp^-1(k) at the end of C

        x = n
        y = n + 1
        xx = n + 2
        yy = n + 3

        ep[x] = y
        ep[y] = x
        ep[xx] = yy
        ep[yy] = xx

        fl[x] = fl[y] = fl[xx] = fl[yy] = fl[i]
        vl[x] = vl[k]
        vl[xx] = vl[y] = vl[j]
        vl[yy] = vl[i]

        fd[fl[i]] += 4
        vd[vl[i]] += 1
        vd[vl[j]] += 2
        vd[vl[k]] += 1

        if i == j == k:
            # face: -> (x xx y yy j C) 
            # (j C kk)
            vp[x] = vp[j]
            vp[yy] = x
            vp[y] = yy
            vp[xx] = y
            vp[j] = xx

            fp[kk] = x
            fp[x] = xx
            fp[xx] = y
            fp[y] = yy
            fp[yy] = j
        elif i == j:
            # face: -> (x xx y k B yy j C)
            # (j C kk) (k B ii)
            vp[yy] = vp[j]
            vp[y] = yy
            vp[xx] = y
            vp[j] = xx
            vp[x] = vp[k]
            vp[k] = x

            fp[ii] = yy
            fp[yy] = j
            fp[kk] = x
            fp[x] = xx
            fp[xx] = y
            fp[y] = k
        elif j == k:
            # face: -> (x xx i A y k B yy)
            # (i A jj) (k B ii)
            vp[yy] = vp[i]
            vp[i] = yy
            vp[y] = vp[k]
            vp[xx] = y
            vp[x] = xx
            vp[k] = x

            fp[ii] = yy
            fp[yy] = x
            fp[x] = xx
            fp[xx] = i
            fp[jj] = y
            fp[y] = k
        elif k == i:
            # face: -> (x xx i A y yy j C)
            # (i A jj) (j C kk)
            vp[y] = vp[j]
            vp[xx] = y
            vp[j] = xx
            vp[x] = vp[i]
            vp[yy] = x
            vp[i] = yy

            fp[kk] = x
            fp[x] = xx
            fp[xx] = i
            fp[jj] = y
            fp[y] = yy
            fp[yy] = j
        else:
            # general case
            # vertex: (...i...)(...j...)(...k...) -> (...i yy...)(...j xx y...)(...k x...)
            # edge  : add (x y) (xx yy)
            # face  : (i A j C k B) -> (x xx i A y k B yy j C)
            #
            # (i A jj) (j C kk) (k B ii)
            vp[yy] = vp[i]
            vp[i] = yy
            vp[y] = vp[j]
            vp[xx] = y
            vp[j] = xx
            vp[x] = vp[k]
            vp[k] = x

            fp[kk] = x
            fp[x] = xx
            fp[xx] = i
            fp[jj] = y
            fp[y] = k
            fp[ii] = yy
            fp[yy] = j

    def remove_face_trisection(self, x):
        ep = self._ep
        fp = self._fp
        fl = self._fl
        fd = self._fd
        vp = self._vp
        vl = self._vl
        vd = self._vd

        x = int(x)
        xx = fp[x]
        y = ep[x]
        yy = ep[xx]

        if fl[x] != fl[y] or fl[x] != fl[xx] or fl[x] != fl[yy]:
            raise ValueError("not a trisection")

        # face: (x xx i A y k B yy j C) -> (i A j C k B)
        #    -> (i A jj) (j C kk) (k B ii)
        i = fp[xx]
        k = fp[y]
        j = fp[yy]
        ii = ep[vp[yy]] # = fp^-1(yy)
        jj = ep[vp[y]]  # = fp^-1(y)
        kk = ep[vp[x]]  # = fp^-1(x)

        if fp[xx] == y and fp[y] == yy:
            # vertex (... j xx y yy x ...) -> (... j ...)
            # face (x xx y yy j C) -> (j C)
            # (j C kk)
            assert vp[j] == xx
            assert vp[xx] == y
            assert vp[yy] == x
            vp[j] = vp[x]
            fp[kk] = j
        elif fp[xx] == y:
            # face: (x xx y k B yy j C) -> (j C k B)
            # (j C kk) (k B ii)
            assert fp[y] != yy and fp[yy] != x
            assert vp[j] == xx
            assert vp[xx] == y
            assert vp[y] == yy
            vp[j] = vp[yy]
            assert vp[k] == x
            vp[k] = vp[x]
            fp[kk] = k
            fp[ii] = j
        elif fp[yy] == x:
            # face: (x xx i A y k B yy) -> (i A k B)
            # (i A jj) (k B ii)
            assert fp[xx] != y and fp[y] != yy
            assert vp[i] == yy
            vp[i] = vp[yy]
            assert vp[k] == x
            assert vp[x] == xx
            assert vp[xx] == y
            vp[k] = vp[y]
            fp[ii] = i
            fp[jj] = k
        elif fp[y] == yy:
            # face: (x xx i A y yy j C) -> (i A j C)
            # (i A jj) (j C kk)
            assert fp[xx] != y and fp[yy] != x
            assert vp[i] == yy
            assert vp[yy] == x
            vp[i] = vp[x]
            assert vp[j] == xx
            assert vp[xx] == y
            vp[j] = vp[y]
            fp[kk] = i
            fp[jj] = j
        else:
            # face: (x xx i A y k B yy j C) -> (i A j C k B)
            # (i A jj) (j C kk) (k B ii)
            assert fp[xx] != y and fp[y] != yy and fp[yy] != x
            assert vp[i] == yy
            vp[i] = vp[yy]
            assert vp[j] == xx
            assert vp[xx] == y
            vp[j] = vp[y]
            assert vp[k] == x
            vp[k] = vp[x]
            fp[jj] = j
            fp[kk] = k
            fp[ii] = i

        self._n -= 4
        fd[fl[i]] -= 4
        vd[vl[i]] -= 1
        vd[vl[j]] -= 2
        vd[vl[k]] -= 1

        assert perm_check(vp, self._n), vp
        assert perm_check(fp, self._n), fp

    ######################################
    # canonical labels and automorphisms #
    ######################################
    # This is chosen so that augmentation works fast (for the exhaustive generation)
    #   - augment1: trisection (single vertex, single face maps)
    #   - augment2: face split (single vertex)
    #   - augment3: vertex split

    def _good_starts(self, i0=-1):
        r"""
        EXAMPLES::

            sage: from surface_dynamics.topology.fat_graph import FatGraph

            sage: CM = []
            sage: w = [0,1,2,3,4,5,0,6,7,1,2,5,3,4,6,7]
            sage: CM.append(FatGraph.from_unicellular_word(w))
            sage: vp = '(0,15,3,6,8)(1,14,18,10,9,12,5,19)(2,7,4,17,11)(13,16)'
            sage: ep = '(0,1)(2,3)(4,5)(6,7)(8,9)(10,11)(12,13)(14,15)(16,17)(18,19)'
            sage: fp = '(0,19,14)(1,8,10,17,13,9,6,2,15)(3,11,18,5,7)(4,12,16)'
            sage: CM.append(FatGraph(vp, ep, fp))
            sage: for cm in CM:
            ....:     gs = set(cm._good_starts())
            ....:     assert gs
            ....:     for i in range(cm.num_darts()):
            ....:         ggs = cm._good_starts(i)
            ....:         if i in gs:
            ....:             ggs = cm._good_starts(i)
            ....:             assert ggs and ggs[0] == i and sorted(ggs) == sorted(gs), (gs, ggs, i)
            ....:         else:
            ....:             assert not ggs, (gs, ggs, i)
        """
        n = self._n
        ep = self._ep
        vd = self._vd
        fd = self._fd
        fp = self._fp
        vl = self._vl
        fl = self._fl
        if i0 == -1:
            ans = []
        else:
            ans = [i0]

        if self._nv > 1:
            # consider only edges with distinct start and end.
            # Maximize the degrees of vertices, then whether the
            # adjacent faces are distinct, then the degree of adjacent
            # faces.
            if i0 != -1:
                j0 = ep[i0]
                if vl[i0] == vl[j0]:
                    return None
                best = (vd[vl[i0]], vd[vl[j0]], fl[i0] != fl[j0], fd[fl[i0]], fd[fl[j0]])
            else:
                best = None
            for i in range(self._n):
                if i == i0:
                    continue
                j = ep[i]
                if vl[i] == vl[j]:
                    continue
                cur = (vd[vl[i]], vd[vl[j]], fl[i] != fl[j], fd[fl[i]], fd[fl[j]])
                if best is None:
                    best = cur
                if cur > best:
                    if i0 != -1:
                        return None
                    else:
                        del ans[:]
                        best = cur
                if cur == best:
                    ans.append(i)

        elif self._nf > 1:
            # (we have a single vertex but several faces)
            # consider only edges with distinct faces on their sides.
            # Maximize their degrees.
            if i0 != -1:
                j0 = ep[i0]
                if fl[i0] == fl[j0]:
                    return None
                best = (fd[fl[i0]], fd[fl[j0]])
            else:
                best = None
            for i in range(self._n):
                if i == i0:
                    continue
                j = ep[i]
                if fl[i] == fl[j]:
                    continue
                cur = (fd[fl[i]], fd[fl[j]])
                if best is None:
                    best = cur
                if cur > best:
                    if i0 != -1:
                        return None
                    else:
                        del ans[:]
                        best = cur
                if cur == best:
                    ans.append(i)
        else:
            # (we have a single face and a single vertex)
            # Minimize the face angle between i and ep[i]
            
            # 1. compute the "face angle" between the half edges
            fa = [None] * n
            fa[0] = 0
            i = fp[0]
            j = 1
            while i != 0:
                fa[i] = j
                i = fp[i]
                j += 1

            # 2. first guess
            if i0 != -1:
                j0 = ep[i0]
                best = fa[j0] - fa[i0]
                if best < 0: best += n
            else:
                best = None

            # 3. run accross the edges
            for i in range(self._n):
                if i == i0:
                    continue
                j = ep[i]
                cur = fa[j] - fa[i]
                if cur < 0: cur += n
                if best is None:
                    best = cur
                if cur < best:
                    if i0 != -1:
                        return None
                    else:
                        del ans[:]
                        best = cur
                if cur == best:
                    ans.append(i)

        return ans

    def _canonical_labelling_from(self, i0):
        r"""
        Edges gets relabelled (2i, 2i+1).

        OUTPUT: a triple ``(fc, fd, rel)`` where
        
        - ``fc`` is the list of edges seen along the walk (with respect to the new
          numbering)
        
        - ``fd``: face degrees seen along the walk

        - ``rel``: relabelling map {current labels} -> {canonical labels}

        EXAMPLES::
   
            sage: from surface_dynamics.topology.fat_graph import FatGraph

            sage: vp = '(0,15,3,6,8)(1,14,18,10,9,12,5,19)(2,7,4,17,11)(13,16)'
            sage: ep = '(0,1)(2,3)(4,5)(6,7)(8,9)(10,11)(12,13)(14,15)(16,17)(18,19)'
            sage: fp = '(0,19,14)(1,8,10,17,13,9,6,2,15)(3,11,18,5,7)(4,12,16)'
            sage: cm = FatGraph(vp, ep, fp)
            sage: for i in range(20):
            ....:     fc, fd, rel = cm._canonical_labelling_from(i)
            ....:     assert len(fc) == 20
            ....:     assert sorted(fd, reverse=True) == [9, 5, 3, 3]
            ....:     assert sorted(rel) == list(range(20))
        """
        n = self._n
        ep = self._ep
        fp = self._fp

        fc = []        # faces seen along the walk
        fd = []        # face degrees seen along the walk
        rel = [-1] * n # dart relabeling

        rel[i0] = 0   # first edge is relabelled (0,1)
        fc.append(0)
        c = 2         # current dart number (in the new labelling scheme)

        # walk along faces first, starting from i0
        # along the way, we collect unseen edges
        i = fp[i0]
        wait = deque([ep[i0]])
        cyc = [0]
        d = 1
        while i != i0:
            if rel[i] == -1:
                j = ep[i]
                if rel[j] != -1:
                    assert rel[j] % 2 == 0
                    rel[i] = rel[j] + 1
                else:
                    rel[i] = c
                    c += 2
                    wait.append(j)
            fc.append(rel[i])
            d += 1
            i = fp[i]
        fd.append(d)

        while wait:
            i0 = wait.popleft()
            if rel[i0] != -1:
                continue
            assert rel[ep[i0]] != -1 and rel[ep[i0]] % 2 == 0
            rel[i0] = rel[ep[i0]] + 1
            fc.append(rel[i0])
            i = fp[i0]
            d = 1
            while i != i0:
                if rel[i] == -1:
                    j = ep[i]
                    if rel[j] != -1:
                        assert rel[j] % 2 == 0
                        rel[i] = rel[j] + 1
                    else:
                        rel[i] = c
                        c += 2
                        wait.append(j)
                fc.append(rel[i])
                i = fp[i]
                d += 1
            fd.append(d)

        assert len(fc) == self._n, (fc, fd, rel)
        assert len(fd) == self._nf, (fc, fd, rel)

        return fc, fd, rel

    # TODO: this is a waste! We should implement the proper partial relabeling
    def _canonical_labelling_from_if_better(self, best, i0):
        r"""
        INPUT:

        - ``best`` - a triple ``(fc, fd, rel)`` as given from _canonical_labeling_from

        - ``i0`` - start edge
        """
        fc_best, fd_best, rel_best = best

        n = self._n
        ep = self._ep
        fp = self._fp
        fl = self._fl
        sfd = self._fd

        is_fd_better = 0  # whether the current face degree works better
        is_fc_better = 0  # whether the current relabelling works better
                          #  0 = equal
                          #  1 = better
                          # -1 = worse

        fd = []        # face degrees seen along the walk (want to maximize)
        fc = []        # edges seen along the walk (want to minimize)
        rel = [-1] * n # dart relabeling

        # walk along faces first, starting from i0
        # along the way, we collect unseen edges

        rel[i0] = 0   # first edge is relabelled (0,1)
        c = 2         # current dart number (in the new labelling scheme)

        d = sfd[fl[i0]]
        if d < fd_best[0]:
            return -1, None
        elif d > fd_best[0]:
            is_fd_better = 1
        fd.append(d)

        fc.append(0)
        i = fp[i0]
        wait = deque([ep[i0]])
        cyc = [0]
        while i != i0:
            if rel[i] == -1:
                j = ep[i]
                if rel[j] != -1:
                    assert rel[j] % 2 == 0
                    rel[i] = rel[j] + 1
                else:
                    rel[i] = c
                    c += 2
                    wait.append(j)

            # edge comparison
            ii = rel[i]
            if not is_fd_better and not is_fc_better:
                if ii < fc_best[len(fc)]:
                    is_fc_better = 1
                elif ii > fc_best[len(fc)]:
                    if self._nf == 1:
                        return -1, None
                    is_fc_better = -1
            fc.append(ii)
            i = fp[i]

        while wait:
            i0 = wait.popleft()

            # face already seen?
            if rel[i0] != -1:
                continue

            # face degree comparison
            d = sfd[fl[i0]]
            if not is_fd_better:
                if d < fd_best[len(fd)]:
                    return -1, None
                elif d > fd_best[len(fd)]:
                    is_fd_better = 1
            fd.append(d)

            # root edge comparison
            assert rel[ep[i0]] != -1 and rel[ep[i0]] % 2 == 0
            ii0 = rel[ep[i0]] + 1
            if not is_fd_better and not is_fc_better:
                if ii0 < fc_best[len(fc)]:
                    is_fc_better = 1
                elif ii0 > fc_best[len(fc)]:
                    is_fc_better = -1

            rel[i0] = ii0
            fc.append(ii0)
            i = fp[i0]

            while i != i0:
                # label the i-th edge (if not already)
                if rel[i] == -1:
                    j = ep[i]
                    if rel[j] != -1:
                        assert rel[j] % 2 == 0
                        rel[i] = rel[j] + 1
                    else:
                        rel[i] = c
                        c += 2
                        wait.append(j)

                # edge comparison
                ii = rel[i]
                if not is_fd_better and not is_fc_better:
                    if ii < fc_best[len(fc)]:
                        is_fc_better = 1
                    elif ii > fc_best[len(fc)]:
                        is_fc_better = -1

                # update
                fc.append(ii)
                i = fp[i]

        cur = (fc, fd, rel)

        if is_fd_better:
            return 1, cur
        else:
            return is_fc_better, cur

    def _is_canonical(self, i0):
        r"""
        Return a pair ``(answer, automorphisms)`` where answer is a boolean
        that says whether this map is in canonical form and ``automorphisms`` form
        a generating set of the group of automorphisms.

        EXAMPLES::

            sage: from surface_dynamics.topology.fat_graph import FatGraph

            sage: vp = '(0,15,3,6,8)(1,14,18,10,9,12,5,19)(2,7,4,17,11)(13,16)'
            sage: ep = '(0,1)(2,3)(4,5)(6,7)(8,9)(10,11)(12,13)(14,15)(16,17)(18,19)'
            sage: fp = '(0,19,14)(1,8,10,17,13,9,6,2,15)(3,11,18,5,7)(4,12,16)'
            sage: cm = FatGraph(vp, ep, fp)
            sage: any(cm._is_canonical(i)[0] for i in range(20))
            True

        A genus 1 example with 4 symmetries::

            sage: vp = '(0,8,6,4,3,7)(1,9,11,5,2,10)'
            sage: ep = '(0,1)(2,3)(4,5)(6,7)(8,9)(10,11)'
            sage: fp = '(0,10,9)(2,4,11)(1,7,8)(3,5,6)'
            sage: cm = FatGraph(vp, ep, fp)
            sage: for i in range(12):
            ....:     test, aut_grp = cm._is_canonical(i)
            ....:     if test: print(aut_grp.group_cardinality())
            4
            4
            4
            4
        """
        roots = self._good_starts(i0)
        if roots is None:
            return False, None

        if len(roots) == 1:
            return True, None

        # perform complete relabelling
        P = PermutationGroupOrbit(self._n, [], roots)
        i = next(P)
        assert i == i0
        best = self._canonical_labelling_from(i)
        rel0 = perm_invert(best[2])

        for i in P:
            test, cur = self._canonical_labelling_from_if_better(best, i)

            if test == 1:
                return False, None

            elif test == 0:
                fc, fd, rel = cur
                aut = perm_compose(rel, rel0)
                P.add_generator(aut)

        return True, P

    def automorphism_group(self):
        r"""
        EXAMPLES::

            sage: from surface_dynamics.topology.fat_graph import FatGraph
            sage: from surface_dynamics.misc.permutation import perm_conjugate

        The four unicellular map with 4 edges in genus 2::

            sage: cm0 = FatGraph.from_unicellular_word([0,1,0,1,2,3,2,3])
            sage: cm1 = FatGraph.from_unicellular_word([0,1,0,2,1,3,2,3])
            sage: cm2 = FatGraph.from_unicellular_word([0,1,0,2,3,1,2,3])
            sage: cm3 = FatGraph.from_unicellular_word([0,1,2,3,0,1,2,3])
            sage: for cm in [cm0, cm1, cm2, cm3]:
            ....:     P = cm.automorphism_group()
            ....:     print(P.group_cardinality())
            ....:     vp = cm.vertex_permutation()
            ....:     ep = cm.edge_permutation()
            ....:     fp = cm.face_permutation()
            ....:     for a in P.gens():
            ....:         pp = perm_conjugate(vp, a)
            ....:         assert pp == vp, (vp, pp)
            2
            1
            1
            8

            sage: cm = FatGraph.from_unicellular_word([0,1,2,3,0,4,1,2,3,4])
            sage: cm.automorphism_group().group_cardinality()
            2

        An example with two faces::

            sage: vp = '(0,9,5,6,7,4,8,1,2,3)'
            sage: ep = '(0,2)(1,3)(4,6)(5,7)(8,9)'
            sage: fp = '(0,1,2,3,8)(4,5,6,7,9)'
            sage: cm = FatGraph(vp,ep,fp)
            sage: cm.automorphism_group()
            PermutationGroupOrbit(10, [(0,4)(1,5)(2,6)(3,7)(8,9)])
        """
        roots = self._good_starts()
        P = PermutationGroupOrbit(self._n, [], roots)
        if len(roots) == 1:
            return P
        i0 = next(P)
        best = self._canonical_labelling_from(i0)
        rel0 = perm_invert(best[2])

        for i in P:
            test, cur = self._canonical_labelling_from_if_better(best, i)

            if test == 1:
                rel0 = perm_invert(cur[2])
                best = cur
            elif test == 0:
                fc, fd, rel = cur
                aut = perm_compose(rel, rel0)
                P.add_generator(aut)

        return P

    def relabel(self, r):
        r"""
        Relabel according to the permutation ``p``

        EXAMPLES::

            sage: from surface_dynamics.topology.fat_graph import FatGraph

            sage: cm = FatGraph.from_unicellular_word([0,1,0,1,2,3,2,3])
            sage: cm.relabel([4,7,0,2,3,5,1,6])
            sage: cm._check()
        """
        n = self._n
        self._vp = perm_conjugate(self._vp, r, n)
        self._ep = perm_conjugate(self._ep, r, n)
        self._fp = perm_conjugate(self._fp, r, n)
        
        vl = [None] * n
        fl = [None] * n
        fa = [None] * n

        for i in range(n):
            j = r[i]
            vl[j] = self._vl[i]
            fl[j] = self._fl[i]

        self._vl = vl
        self._fl = fl
